/* 
 * tclExecute.c --
 *
 *	This file contains procedures that execute byte-compiled Tcl
 *	commands.
 *
 * Copyright (c) 1996-1997 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclExecute.c 1.49 97/01/22 19:03:51
 */

#include "tclInt.h"
#include "tclCompile.h"

#ifdef NO_FLOAT_H
#   include "../compat/float.h"
#else
#   include <float.h>
#endif
#ifndef TCL_NO_MATH
#include <math.h>
#endif

/*
 * The stuff below is a bit of a hack so that this file can be used
 * in environments that include no UNIX, i.e. no errno.  Just define
 * errno here.
 */

#ifndef TCL_GENERIC_ONLY
#include "tclPort.h"
#else
#define NO_ERRNO_H
#endif

#ifdef NO_ERRNO_H
int errno;
#define EDOM 33
#define ERANGE 34
#endif

/*
 * Boolean flag indicating whether the Tcl bytecode interpreter has been
 * initialized.
 */

static int execInitialized = 0;

/*
 * Variable that controls whether execution tracing is enabled and, if so,
 * what level of tracing is desired:
 *    0: no execution tracing
 *    1: trace invocations of Tcl procs only
 *    2: trace invocations of all (not compiled away) commands
 *    3: display each instruction executed
 * This variable is linked to the Tcl variable "traceExec".
 */

int tclTraceExec = 0;

/*
 * The following global variable is use to signal matherr that Tcl
 * is responsible for the arithmetic, so errors can be handled in a
 * fashion appropriate for Tcl.  Zero means no Tcl math is in
 * progress;  non-zero means Tcl is doing math.
 */

int tcl_MathInProgress = 0;

/*
 * The variable below serves no useful purpose except to generate
 * a reference to matherr, so that the Tcl version of matherr is
 * linked in rather than the system version. Without this reference
 * the need for matherr won't be discovered during linking until after
 * libtcl.a has been processed, so Tcl's version won't be used.
 */

#ifdef NEED_MATHERR
extern int matherr();
int (*tclMatherrPtr)() = matherr;
#endif

/*
 * Array of instruction names.
 */

static char *opName[256];

/*
 * Mapping from expression instruction opcodes to strings; used for error
 * messages. Note that these entries must match the order and number of the
 * expression opcodes (e.g., INST_LOR) in tclCompile.h.
 */

static char *operatorStrings[] = {
    "|", "^", "&", "==", "!=", "<", ">", "<=", ">=", "<<", ">>",
    "+", "-", "*", "/", "%", "+", "-", "~", "!",
    "BUILTIN FUNCTION", "FUNCTION"
};
    
/*
 * Mapping from Tcl result codes to strings; used for error and debugging
 * messages. 
 */

#ifdef TCL_COMPILE_DEBUG
static char *resultStrings[] = {
    "TCL_OK", "TCL_ERROR", "TCL_RETURN", "TCL_BREAK", "TCL_CONTINUE"
};
#endif /* TCL_COMPILE_DEBUG */

/*
 * The following are statistics-related variables that record information
 * about the bytecode compiler and interpreter's operation. This includes
 * an array that records for each instruction how often it is executed.
 */

#ifdef TCL_COMPILE_STATS
static int instructionCount[256];
static long numExecutions = 0;
#endif /* TCL_COMPILE_STATS */

/*
 * Macros for testing floating-point values for certain special cases. Test
 * for not-a-number by comparing a value against itself; test for infinity
 * by comparing against the largest floating-point value.
 */

#define IS_NAN(v) ((v) != (v))
#ifdef DBL_MAX
#   define IS_INF(v) (((v) > DBL_MAX) || ((v) < -DBL_MAX))
#else
#   define IS_INF(v) 0
#endif

/*
 * Macro to adjust the program counter and restart the instruction execution
 * loop after each instruction is executed.
 */

#define ADJUST_PC(instBytes) \
    pc += instBytes;  continue

/*
 * Macros used to cache (decache) Tcl evaluation stack information in (from)
 * local variables.
 */

#define CACHE_STACK_INFO() \
    stackPtr = eePtr->stackPtr; \
    stackTop = eePtr->stackTop

#define DECACHE_STACK_INFO() \
    eePtr->stackTop = stackTop

/*
 * Macros used to access items on the Tcl evaluation stack. Note that while
 * PUSH_OBJECT increments the argument object's ref count (since the stack
 * then has another reference pointing to the object), POP_OBJECT does _not_
 * decrement the ref count. This is because the stack may hold the only
 * reference to the object, so the object might be destroyed if its ref
 * count were decremented before you had a chance to, e.g., store it in a
 * variable. It is your responsibility to decrement the ref count when you
 * are finished with an object.
 */

#define STK_ITEM(offset)    (stackPtr[stackTop + (offset)])
#define STK_OBJECT(offset)  (STK_ITEM(offset).o)
#define STK_INT(offset)     (STK_ITEM(offset).i)
#define STK_POINTER(offset) (STK_ITEM(offset).p)

#define PUSH_OBJECT(objPtr) \
    stackTop++; \
    stackPtr[stackTop].o = (objPtr); \
    Tcl_IncrRefCount(objPtr)
    
#define POP_OBJECT() \
    (stackPtr[stackTop--].o)

/*
 * Macros used to trace instruction execution. The macros TRACE,
 * TRACE_WITH_OBJ, and O2S are only used inside TclExecuteByteCode.
 * O2S is only used in TRACE* calls to get a string from an object.
 * 
 * NOTE THAT CLIENTS OF O2S ARE LIKELY TO FAIL IF THE OBJECT'S
 * STRING REP CONTAINS NULLS. 
 */

#ifdef TCL_COMPILE_DEBUG
	
#define O2S(objPtr) \
    TclGetStringFromObj((objPtr), &length)
	
#ifdef TCL_COMPILE_STATS
#define TRACE(a) \
    if (traceInstructions) { \
        fprintf(stdout, "%d: %d,%ld (%u) ", iPtr->numLevels, \
	       stackTop, (tclObjsAlloced - tclObjsFreed), \
	       (unsigned int)(pc - codePtr->codeStart)); \
	printf a; \
        fflush(stdout); \
    }
#define TRACE_WITH_OBJ(a, objPtr) \
    if (traceInstructions) { \
        fprintf(stdout, "%d: %d,%ld (%u) ", iPtr->numLevels, \
	       stackTop, (tclObjsAlloced - tclObjsFreed), \
	       (unsigned int)(pc - codePtr->codeStart)); \
	printf a; \
	bytes = TclGetStringFromObj((objPtr), &length); \
        TclPrintSource(stdout, bytes, TclMin(length, 30)); \
        fprintf(stdout, "\n"); \
        fflush(stdout); \
    }
#else  /* not TCL_COMPILE_STATS */
#define TRACE(a) \
    if (traceInstructions) { \
        fprintf(stdout, "%d: %d (%u) ", iPtr->numLevels, stackTop, \
	       (unsigned int)(pc - codePtr->codeStart)); \
	printf a; \
        fflush(stdout); \
    }
#define TRACE_WITH_OBJ(a, objPtr) \
    if (traceInstructions) { \
        fprintf(stdout, "%d: %d (%u) ", iPtr->numLevels, stackTop, \
	       (unsigned int)(pc - codePtr->codeStart)); \
	printf a; \
	bytes = TclGetStringFromObj((objPtr), &length); \
        TclPrintSource(stdout, bytes, TclMin(length, 30)); \
        fprintf(stdout, "\n"); \
        fflush(stdout); \
    }
#endif /* TCL_COMPILE_STATS */

#else  /* not TCL_COMPILE_DEBUG */
	
#define TRACE(a)
#define TRACE_WITH_OBJ(a, objPtr)
#define O2S(objPtr)
	
#endif /* TCL_COMPILE_DEBUG */

/*
 * Declarations for local procedures to this file:
 */

static void		CallTraceProcedure _ANSI_ARGS_((Tcl_Interp *interp,
			    Trace *tracePtr, Command *cmdPtr,
			    int objc, Tcl_Obj *objv[]));
static void		DupCmdNameInternalRep _ANSI_ARGS_((Tcl_Obj *objPtr,
			    Tcl_Obj *copyPtr));
static int		ExprAbsFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
static int		ExprBinaryFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
static int		ExprCallMathFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, int objc, Tcl_Obj **objv));
static int		ExprDoubleFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
static int		ExprIntFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
static int		ExprRoundFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
static int		ExprUnaryFunc _ANSI_ARGS_((Tcl_Interp *interp,
			    ExecEnv *eePtr, ClientData clientData));
#ifdef TCL_COMPILE_STATS
static int              EvalStatsCmd _ANSI_ARGS_((ClientData clientData,
                            Tcl_Interp *interp, int argc, char **argv));
#endif /* TCL_COMPILE_STATS */
static void		FreeCmdNameInternalRep _ANSI_ARGS_((
    			    Tcl_Obj *objPtr));
static void		GrowEvaluationStack _ANSI_ARGS_((ExecEnv *eePtr));
static void		IllegalExprOperandType _ANSI_ARGS_((
			    Tcl_Interp *interp, unsigned int opCode,
			    Tcl_Obj *opndPtr));
static void		InitByteCodeExecution _ANSI_ARGS_((
			    Tcl_Interp *interp));
static int		SetCmdNameFromAny _ANSI_ARGS_((Tcl_Interp *interp,
			    Tcl_Obj *objPtr));
#ifdef TCL_COMPILE_DEBUG
static char *		StringForResultCode _ANSI_ARGS_((int result));
#endif /* TCL_COMPILE_DEBUG */
static void		UpdateStringOfCmdName _ANSI_ARGS_((Tcl_Obj *objPtr));

/*
 * Table describing the built-in math functions. Entries in this table are
 * indexed by the values of the INST_CALL_BUILTIN_FUNC instruction's
 * operand byte.
 */

BuiltinFunc builtinFuncTable[] = {
#ifndef TCL_NO_MATH
    {"acos", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) acos},
    {"asin", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) asin},
    {"atan", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) atan},
    {"atan2", 2, {TCL_DOUBLE, TCL_DOUBLE}, ExprBinaryFunc, (ClientData) atan2},
    {"ceil", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) ceil},
    {"cos", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) cos},
    {"cosh", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) cosh},
    {"exp", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) exp},
    {"floor", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) floor},
    {"fmod", 2, {TCL_DOUBLE, TCL_DOUBLE}, ExprBinaryFunc, (ClientData) fmod},
    {"hypot", 2, {TCL_DOUBLE, TCL_DOUBLE}, ExprBinaryFunc, (ClientData) hypot},
    {"log", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) log},
    {"log10", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) log10},
    {"pow", 2, {TCL_DOUBLE, TCL_DOUBLE}, ExprBinaryFunc, (ClientData) pow},
    {"sin", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) sin},
    {"sinh", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) sinh},
    {"sqrt", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) sqrt},
    {"tan", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) tan},
    {"tanh", 1, {TCL_DOUBLE}, ExprUnaryFunc, (ClientData) tanh},
#endif
    {"abs", 1, {TCL_EITHER}, ExprAbsFunc, 0},
    {"double", 1, {TCL_EITHER}, ExprDoubleFunc, 0},
    {"int", 1, {TCL_EITHER}, ExprIntFunc, 0},
    {"round", 1, {TCL_EITHER}, ExprRoundFunc, 0},
    {0},
};

/*
 * The structure below defines the command name Tcl object type by means of
 * procedures that can be invoked by generic object code. Objects of this
 * type cache the Command pointer that results from looking up command names
 * in the command hashtable. Such objects appear as the zeroth ("command
 * name") argument in a Tcl command.
 */

Tcl_ObjType tclCmdNameType = {
    "cmdName",				/* name */
    FreeCmdNameInternalRep,		/* freeIntRepProc */
    DupCmdNameInternalRep,		/* dupIntRepProc */
    UpdateStringOfCmdName,		/* updateStringProc */
    SetCmdNameFromAny			/* setFromAnyProc */
};

/*
 *----------------------------------------------------------------------
 *
 * InitByteCodeExecution --
 *
 *	This procedure is called once to initialize the Tcl bytecode
 *	interpreter.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	This procedure initializes the array of instruction names. If
 *	compiling with the TCL_COMPILE_STATS flag, it initializes the
 *	array that counts the executions of each instruction and it
 *	creates the "evalstats" command. It also registers the command name
 *	Tcl_ObjType. It also establishes the link between the Tcl
 *	"traceExec" and C "tclTraceExec" variables.
 *
 *----------------------------------------------------------------------
 */

static void
InitByteCodeExecution(interp)
    Tcl_Interp *interp;		/* Interpreter for which the Tcl variable
				 * "traceExec" is linked to control
				 * instruction tracing. */
{
    int i;
    
    Tcl_RegisterObjType(&tclCmdNameType);

    (VOID *) memset(opName, 0, sizeof(opName));
    for (i = 0;  instructionTable[i].name != NULL;  i++) {
	opName[i] = instructionTable[i].name;
    }

#ifdef TCL_COMPILE_STATS    
    (VOID *) memset(instructionCount, 0, sizeof(instructionCount));
    Tcl_CreateCommand(interp, "evalstats", EvalStatsCmd,
		      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
#endif /* TCL_COMPILE_STATS */
    
    if (Tcl_LinkVar(interp, "traceExec", (char *) &tclTraceExec,
		    TCL_LINK_INT) != TCL_OK) {
	panic("InitByteCodeExecution: can't create link for traceExec variable");
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TclCreateExecEnv --
 *
 *	This procedure creates a new execution environment for Tcl bytecode
 *	execution. An ExecEnv points to a Tcl evaluation stack. An ExecEnv
 *	is typically created once for each Tcl interpreter (Interp
 *	structure) and recursively passed to TclExecuteByteCode to execute
 *	ByteCode sequences for nested commands.
 *
 * Results:
 *	A newly allocated ExecEnv is returned. This points to an empty
 *	evaluation stack of the standard initial size.
 *
 * Side effects:
 *	The bytecode interpreter is also initialized here, as this
 *	procedure will be called before any call to TclExecuteByteCode.
 *
 *----------------------------------------------------------------------
 */

#define TCL_STACK_INITIAL_SIZE 2000

ExecEnv *
TclCreateExecEnv(interp)
    Tcl_Interp *interp;		/* Interpreter for which the execution
				 * environment is being created. */
{
    ExecEnv *eePtr = (ExecEnv *) ckalloc(sizeof(ExecEnv));

    eePtr->stackPtr = (StackItem *)
	ckalloc((unsigned) (TCL_STACK_INITIAL_SIZE * sizeof(StackItem)));
    eePtr->stackTop = -1;
    eePtr->stackEnd = (TCL_STACK_INITIAL_SIZE - 1);

    if (!execInitialized) {
	InitByteCodeExecution(interp);
	execInitialized = 1;
    }

    return eePtr;
}
#undef TCL_STACK_INITIAL_SIZE

/*
 *----------------------------------------------------------------------
 *
 * TclDeleteExecEnv --
 *
 *	Frees the storage for an ExecEnv.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Storage for an ExecEnv and its contained storage (e.g. the
 *	evaluation stack) is freed.
 *
 *----------------------------------------------------------------------
 */

void
TclDeleteExecEnv(eePtr)
    ExecEnv *eePtr;		/* Execution environment to free. */
{
    ckfree((char *) eePtr->stackPtr);
    ckfree((char *) eePtr);
}

/*
 *----------------------------------------------------------------------
 *
 * GrowEvaluationStack --
 *
 *	This procedure grows a Tcl evaluation stack stored in an ExecEnv.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The size of the evaluation stack is doubled.
 *
 *----------------------------------------------------------------------
 */

static void
GrowEvaluationStack(eePtr)
    register ExecEnv *eePtr; /* Points to the ExecEnv with an evaluation
			      * stack to enlarge. */
{
    /*
     * The current Tcl stack elements are stored from eePtr->stackPtr[0]
     * to eePtr->stackPtr[eePtr->stackEnd] (inclusive).
     */

    int currElems = (eePtr->stackEnd + 1);
    int newElems  = 2*currElems;
    int currBytes = currElems * sizeof(StackItem);
    int newBytes  = 2*currBytes;
    StackItem *newStackPtr = (StackItem *) ckalloc((unsigned) newBytes);

    /*
     * Copy the existing stack items to the new stack space, free the old
     * storage if appropriate, and mark new space as malloc'ed.
     */
 
    memcpy((VOID *) newStackPtr, (VOID *) eePtr->stackPtr,
	   (size_t) currBytes);
    ckfree((char *) eePtr->stackPtr);
    eePtr->stackPtr = newStackPtr;
    eePtr->stackEnd = (newElems - 1); /* i.e. index of last usable item */
}

/*
 *----------------------------------------------------------------------
 *
 * TclExecuteByteCode --
 *
 *	This procedure executes the instructions of a ByteCode structure.
 *	It returns when a "done" instruction is executed or an error occurs.
 *
 * Results:
 *	The return value is one of the return codes defined in tcl.h
 *	(such as TCL_OK), and interp->objResult refers to a Tcl object that
 *	contains the result of executing the bytecode or an error message,
 *
 * Side effects:
 *	Almost certainly, depending on the ByteCode's instructions.
 *
 *----------------------------------------------------------------------
 */

int
TclExecuteByteCode(interp, codePtr)
    Tcl_Interp *interp;		/* Token for command interpreter. */
    ByteCode *codePtr;		/* The bytecode sequence to interpret. */
{
    Interp *iPtr = (Interp *) interp;
    ExecEnv *eePtr = iPtr->execEnvPtr;
    				/* Points to the execution environment. */
    register StackItem *stackPtr = eePtr->stackPtr;
    				/* Cached evaluation stack base pointer. */
    register int stackTop = eePtr->stackTop;
    				/* Cached top index of evaluation stack. */
    Tcl_Obj **objArrayPtr = codePtr->objArrayPtr;
    				/* Points to the ByteCode's object array. */
    unsigned char *pc = codePtr->codeStart;
				/* The current program counter. */
    unsigned char opCode;	/* The current instruction code. */
    int opnd;			/* Current instruction's operand byte. */
    int pcAdjustment;		/* Hold pc adjustment after instruction. */
    int initStackTop = stackTop;/* Stack top at start of execution. */
    ExceptionRange *rangePtr;	/* Points to closest loop or catch exception
				 * range enclosing the pc. Used by various
				 * instructions and processCatch to
				 * process break, continue, and errors. */
    int result = TCL_OK;	/* Return code returned after execution. */
    int traceInstructions = (tclTraceExec == 3);
    Tcl_Obj *valuePtr, *value2Ptr, *namePtr, *objPtr;
    char *bytes;
    int length;
    long i;
    Tcl_DString command;	/* Used for debugging. If tclTraceExec >= 2
				 * holds a string representing the last
				 * command invoked. */

    /*
     * This procedure uses a stack to hold information about catch commands.
     * This information is the current operand stack top when starting to
     * execute the code for each catch command. It starts out with stack-
     * allocated space but uses dynamically-allocated storage if needed.
     */

#define STATIC_CATCH_STACK_SIZE 5
    int (catchStackStorage[STATIC_CATCH_STACK_SIZE]);
    int *catchStackPtr = catchStackStorage;
    int catchTop = -1;

    /*
     * THIS PROC FAILS IF AN OBJECT'S STRING REP HAS A NULL BYTE.
     */
    
    if (tclTraceExec >= 2) {
	Proc *procPtr = codePtr->procPtr;
	fprintf(stdout, "\nExecuting ByteCode 0x%x, epoch=%u, commands=%d\n",
	        (unsigned int) codePtr, codePtr->compileEpoch,
	        codePtr->numCommands);
	if (procPtr != NULL) {
	    fprintf(stdout,
	            "  Proc 0x%x, ref ct=%d, %d args, %d compiled locals\n",
		    (unsigned int) procPtr, procPtr->refCount,
		    procPtr->numArgs, procPtr->numCompiledLocals);
	}
	fprintf(stdout, "  Source: ");
	TclPrintSource(stdout, codePtr->source, 70);
	fprintf(stdout, "\n");
	fprintf(stdout, "  Chars=%d, bytes=%u, objs=%u, stk depth=%u, exc depth=%d, aux items=%d\n",
		codePtr->numSrcChars, codePtr->numCodeBytes,
		codePtr->numObjects, codePtr->maxStackDepth,
		codePtr->maxExcRangeDepth, codePtr->numAuxDataItems);
#ifdef TCL_COMPILE_STATS
	fprintf(stdout, "  Starting stack top=%d, system objects=%ld\n",
		eePtr->stackTop, (tclObjsAlloced - tclObjsFreed));
#else
	fprintf(stdout, "  Starting stack top=%d\n", eePtr->stackTop);
#endif /* TCL_COMPILE_STATS */
	fflush(stdout);
    }

#ifdef TCL_COMPILE_STATS
    numExecutions++;
#endif /* TCL_COMPILE_STATS */

    /*
     * Make sure the catch stack is large enough to hold the maximum number
     * of catch commands that could ever be executing at the same time. This
     * will be no more than the exception range array's depth.
     */

    if (codePtr->maxExcRangeDepth > STATIC_CATCH_STACK_SIZE) {
	catchStackPtr = (int *)
	        ckalloc(codePtr->maxExcRangeDepth * sizeof(int));
    }

    /*
     * Make sure the stack has enough room to execute this ByteCode.
     */

    if ((stackTop + codePtr->maxStackDepth) > eePtr->stackEnd) {
        GrowEvaluationStack(eePtr); 
        stackPtr = eePtr->stackPtr;
    }

    /*
     * Initialize the buffer that holds a string containing the name and
     * arguments for the last invoked command.
     */

    Tcl_DStringInit(&command);

    /*
     * Loop executing instructions until a "done" instruction, a TCL_RETURN,
     * or some error.
     */

    for (;;) {
        opCode = *pc;

#ifdef TCL_COMPILE_DEBUG
	if (((unsigned int) pc < (unsigned int) codePtr->codeStart) ||
	    ((unsigned int) pc > (unsigned int) (codePtr->codeStart + codePtr->numCodeBytes))) {
	    fprintf(stderr, "\nTclExecuteByteCode: bad instruction pc 0x%x\n",
		    (unsigned int) pc);
	    panic("TclExecuteByteCode execution failure: bad pc");
	}
	if ((unsigned int) opCode > LAST_INST_OPCODE) {
	    fprintf(stderr, "\nTclExecuteByteCode: bad opcode %d at pc %u\n",
		    (unsigned int) opCode, (unsigned int)(pc - codePtr->codeStart));
	    panic("TclExecuteByteCode execution failure: bad opcode");
	}
	if ((stackTop < initStackTop) || (stackTop > eePtr->stackEnd)) {
	    int cmdIndex = TclGetSrcInfoForPc(pc, codePtr);
	    fprintf(stderr, "\nTclExecuteByteCode: bad stack top %d at pc %u",
	            stackTop, (unsigned int)(pc - codePtr->codeStart));
	    if (cmdIndex != -1) {
		CmdLocation *locPtr = &(codePtr->cmdMapPtr[cmdIndex]);
		char *ellipsis = "";
		int numChars = locPtr->numSrcChars;
		if (numChars > 100) {
		    numChars = 100;
		    ellipsis = "...";
		}
		fprintf(stderr, "\n executing %.*s%s\n", numChars,
			(codePtr->source + locPtr->srcOffset), ellipsis);
	    } else {
		fprintf(stderr, "\n");
	    }
	    panic("TclExecuteByteCode execution failure: bad stack top");
	}
#else /* not TCL_COMPILE_DEBUG - print generic trace if so requested */
	if (traceInstructions) {
#ifdef TCL_COMPILE_STATS
	    fprintf(stdout, "%d: %d,%ld ", iPtr->numLevels, stackTop,
		    (tclObjsAlloced - tclObjsFreed));
#else /* TCL_COMPILE_STATS */
	    fprintf(stdout, "%d: %d ", iPtr->numLevels, stackTop);
#endif /* TCL_COMPILE_STATS */
	    TclPrintInstruction(interp, codePtr, pc);
	    fflush(stdout);
	}
#endif /* TCL_COMPILE_DEBUG */
	
#ifdef TCL_COMPILE_STATS    
	instructionCount[opCode]++;
#endif /* TCL_COMPILE_STATS */
	
        switch (opCode) {
	case INST_DONE:
	    /*
	     * Pop the topmost object from the stack, set the interpreter's
	     * object result to point to it, and return. Since the
	     * interpreter now has another reference to it, we don't
	     * decrement the object's ref count after popping it.
	     */
	    valuePtr = iPtr->objResult;
	    TclDecrRefCount(valuePtr);
	    iPtr->objResult = POP_OBJECT();
	    if (stackTop != initStackTop) {
		fprintf(stderr, "\nTclExecuteByteCode: done instruction at pc %u: stack top %d != entry stack top %d\n",
			(unsigned int)(pc - codePtr->codeStart),
			(unsigned int) stackTop,
			(unsigned int) initStackTop);
		fprintf(stderr, "  Source: ");
		TclPrintSource(stderr, codePtr->source, 150);
		panic("TclExecuteByteCode execution failure: end stack top != start stack top");
	    }
	    TRACE_WITH_OBJ(("done => return code=%d, result is ", result),
		    iPtr->objResult);
	    goto done;
	    
	case INST_PUSH1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    valuePtr = objArrayPtr[opnd];
	    stackPtr[++stackTop].o = valuePtr;
	    Tcl_IncrRefCount(valuePtr);
	    TRACE_WITH_OBJ(("push1 %u => ", opnd), valuePtr);
	    ADJUST_PC(2);
	    
	case INST_PUSH4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    valuePtr = objArrayPtr[opnd];
	    PUSH_OBJECT(valuePtr);
	    TRACE_WITH_OBJ(("push4 %u => ", opnd), valuePtr);
	    ADJUST_PC(5);
	    
	case INST_POP:
	    valuePtr = POP_OBJECT();
	    TRACE_WITH_OBJ(("pop => discarding "), valuePtr);
	    TclDecrRefCount(valuePtr); /* finished with pop'ed object. */
	    ADJUST_PC(1);

	case INST_DUP:
	    valuePtr = stackPtr[stackTop].o;
	    value2Ptr = Tcl_DuplicateObj(valuePtr);
	    stackPtr[++stackTop].o = value2Ptr; /* already has ref ct==1 */
	    TRACE_WITH_OBJ(("dup => "), value2Ptr);
	    ADJUST_PC(1);

	case INST_CONCAT1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    {
		Tcl_Obj *concatObjPtr;
		int totalLen = 0;

		/*
		 * Concatenate strings from the top opnd items on the stack
		 * starting with the deepest item. First, determine how many
		 * characters are needed.
		 */

		for (i = (stackTop - (opnd-1));  i <= stackTop;  i++) {
		    valuePtr = stackPtr[i].o;
		    bytes = TclGetStringFromObj(valuePtr, &length);
		    if (bytes != NULL) {
			totalLen += length;
		    }
                }

		/*
		 * Initialize the new append string object by appending the
		 * strings of the opnd stack objects. Also pop the objects. 
		 */

		TclNewObj(concatObjPtr);
		if (totalLen > 0) {
		    char *p = (char *) ckalloc((unsigned) (totalLen + 1));
		    
		    concatObjPtr->bytes = p;
		    concatObjPtr->length = totalLen;

		    for (i = (stackTop - (opnd-1));  i <= stackTop;  i++) {
			valuePtr = stackPtr[i].o;
			bytes = TclGetStringFromObj(valuePtr, &length);
			if (bytes != NULL) {
			    memcpy((VOID *) p, (VOID *) bytes,
				    (size_t) length);
			    p += length;
			}
			TclDecrRefCount(valuePtr);
		    }
		    *p = '\0';
		    stackTop -= opnd;
		} else {
		    i = (stackTop - (opnd-1));
		    while (i <= stackTop) {
			valuePtr = stackPtr[i].o;
			TclDecrRefCount(valuePtr);
			i++;
		    }
		    stackTop -= opnd;
		}
		
		/*
		 * Push the result object which already has ref count 1.
		 */

		stackPtr[++stackTop].o = concatObjPtr;
		TRACE_WITH_OBJ(("concat %u => ", opnd), concatObjPtr);
		ADJUST_PC(2);
            }
	    
	case INST_INVOKE_STK4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doInvocation;

	case INST_INVOKE_STK1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doInvocation:
	    {
		char *cmdName;
		Command *cmdPtr;   /* Points to command's Command struct. */
		int objc = opnd;   /* The number of arguments. */
		Tcl_Obj **objv;	   /* The array of argument objects. */
		Tcl_Obj *objv0Ptr; /* Holds objv[0], the command name. */
		int newPcOffset = 0;
				   /* Instruction offset computed during
				    * break, continue, error processing.
				    * Init. to avoid compiler warning. */
		Trace *tracePtr;
		Tcl_HashEntry *hPtr, *cachedHtblPtr;
#ifdef TCL_COMPILE_DEBUG
		int isUnknownCmd = 0;
		char cmdNameBuf[30];
#endif /* TCL_COMPILE_DEBUG */
		
		/*
		 * If the interpreter was deleted, return an error.
		 */
		
		if (iPtr->flags & DELETED) {
		    Tcl_ResetResult(interp);
		    interp->result =
			    "attempt to call eval in deleted interpreter";
		    TclSetObjResultFromStrResult(interp);
		    Tcl_SetErrorCode(interp, "CORE", "IDELETE",
				     interp->result, (char *) NULL);
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
    
		objv = &(stackPtr[stackTop - (objc-1)].o);
		objv0Ptr = objv[0];
		cmdName = TclGetStringFromObj(objv0Ptr, &length);
		
		/*
		 * Find the procedure to execute this command. If there
		 * isn't one, then see if there is a command "unknown". If
		 * so, invoke it, passing it the original command words as
		 * arguments.
		 *
		 * We convert the objv[0] object to be a CmdName object.
		 * This caches a pointer to the Command structure for the
		 * command. It also caches a pointer to the object-based
		 * command procedure that we last found in the Command
		 * structure. If the Command structure pointer is NULL,
		 * or if the command procedure is NULL (i.e., the command
		 * was deleted) or the hashtable pointer has changed
		 * (indicating that the command was renamed or was made
		 * hidden), our cached values are invalid and we must do
		 * a hashtable lookup.
		 * cmdName FAILS IF IT HAS EMBEDDED NULLs.
		 */

		if (objv0Ptr->typePtr != &tclCmdNameType) {
		    SetCmdNameFromAny(interp, objv0Ptr);
		}
		cmdPtr = (Command *)
		    objv0Ptr->internalRep.twoPtrValue.ptr1;
		cachedHtblPtr = (Tcl_HashEntry *)
		    objv0Ptr->internalRep.twoPtrValue.ptr2;

		if ((cmdPtr == NULL)
		        || (cmdPtr->objProc == NULL)
		        || (cmdPtr->hPtr != cachedHtblPtr)) {
		    if (cmdPtr != NULL) {
			/*
			 * The cached Command pointer has become invalid.
			 * Decrement the old Command structure's ref count
			 * and free it if we had the last reference to it.
			 */
			TclCleanupCommand(cmdPtr);
			objv0Ptr->typePtr = NULL;
			cmdPtr = NULL;
		    }
		    hPtr = Tcl_FindHashEntry(&iPtr->commandTable, cmdName);
		    if (hPtr != NULL) {
			cmdPtr = (Command *) Tcl_GetHashValue(hPtr);
		    }
		    if (cmdPtr != NULL) {   /* update the cached values */
		        objv0Ptr->internalRep.twoPtrValue.ptr1 =
			    (VOID *) cmdPtr;
		        objv0Ptr->internalRep.twoPtrValue.ptr2 =
			    (VOID *) hPtr;
			objv0Ptr->typePtr = &tclCmdNameType;
			cmdPtr->refCount++; /* we have another reference */
		    } else {
			/*
			 * The command was still not found.
			 */
			hPtr = Tcl_FindHashEntry(&iPtr->commandTable,
						 "unknown");
			if (hPtr == NULL) {
			    Tcl_ResetResult(interp);
			    Tcl_AppendResult(interp,
			            "invalid command name \"", cmdName,
				    "\"", (char *) NULL);
			    TRACE(("%s %u => unknown proc not found: ",
				    opName[opCode], objc));
			    TclSetObjResultFromStrResult(interp);
			    result = TCL_ERROR;
			    goto checkForCatch;
			}
			cmdPtr = (Command *) Tcl_GetHashValue(hPtr);
#ifdef TCL_COMPILE_DEBUG
			isUnknownCmd = 1;
#endif /*TCL_COMPILE_DEBUG*/			
			stackTop++; /* need room for new inserted objv[0] */
			for (i = objc;  i >= 0;  i--) {
			    objv[i+1] = objv[i];
			}
			objc++;
			
			TclNewObj(objv[0]);
			objv[0]->bytes = (char *) ckalloc(8);
			strcpy(objv[0]->bytes, "unknown");
			objv[0]->length = 7;
		    }
		}
		
		/*
		 * Call any trace procedures.
		 */
		
		for (tracePtr = iPtr->tracePtr;  tracePtr != NULL;
		         tracePtr = tracePtr->nextPtr) {
		    if (iPtr->numLevels <= tracePtr->level) {
			CallTraceProcedure(interp, tracePtr, cmdPtr,
					   objc, objv);
		    }
		}
		
		/*
		 * Finally, invoke the command's Tcl_ObjCmdProc. First reset
		 * the interpreter's string and object results to their
		 * default empty values since they could have gotten changed
		 * by earlier invocations.
		 */
		
		TclResetObjResult(iPtr);
		Tcl_FreeResult(iPtr);
		iPtr->result = iPtr->resultSpace;
		iPtr->resultSpace[0] = 0;
		iPtr->flags &= ~(ERR_ALREADY_LOGGED | ERR_IN_PROGRESS
                        | ERROR_CODE_SET);

		if (tclTraceExec >= 2) {
		    char buffer[50];

		    sprintf(buffer, "%d: (%u) invoking ", iPtr->numLevels,
			    (unsigned int)(pc - codePtr->codeStart));
		    Tcl_DStringAppend(&command, buffer, -1);
		    
#ifdef TCL_COMPILE_DEBUG
		    if (traceInstructions) { /* tclTraceExec == 3 */
			strncpy(cmdNameBuf, cmdName, 20);
			TRACE(("%s %u => call ", opName[opCode],
			       (isUnknownCmd? objc-1 : objc)));
		    } else {
			fprintf(stdout, "%s", buffer);
		    }
#else /* TCL_COMPILE_DEBUG */
		    fprintf(stdout, "%s", buffer);
#endif /*TCL_COMPILE_DEBUG*/

		    for (i = 0;  i < objc;  i++) {
			bytes = TclGetStringFromObj(objv[i], &length);
			TclPrintSource(stdout, bytes, TclMin(length, 15));
			fprintf(stdout, " ");

			sprintf(buffer, "\"%.*s\" ", TclMin(length, 15), bytes);
			Tcl_DStringAppend(&command, buffer, -1);
		    }
		    fprintf(stdout, "\n");
		    fflush(stdout);

		    Tcl_DStringFree(&command);
		}

		iPtr->cmdCount++;
		DECACHE_STACK_INFO();
		result = (*cmdPtr->objProc)(cmdPtr->objClientData, interp,
					    objc, objv);
		CACHE_STACK_INFO();
		if (Tcl_AsyncReady()) {
		    result = Tcl_AsyncInvoke(interp, result);
		}
		
		/*
		 * Pop the objc top stack elements and decrement their ref
		 * counts. 
		 */
		
		i = (stackTop - (objc-1));
		while (i <= stackTop) {
		    valuePtr = stackPtr[i].o;
		    TclDecrRefCount(valuePtr);
		    i++;
		}
		stackTop -= objc;

		/*
		 * Process the result of the Tcl_ObjCmdProc call.
		 */
		
		switch (result) {
		case TCL_OK:
		    /*
		     * Push the call's object result and continue execution
		     * with the next instruction.
		     */
		    valuePtr = iPtr->objResult;
		    Tcl_IncrRefCount(valuePtr);  /* stack points to obj */
		    stackPtr[++stackTop].o = valuePtr;
		    TRACE_WITH_OBJ(("%s %u => ...after \"%.20s\", result=",
		            opName[opCode], objc, cmdNameBuf), valuePtr);
		    ADJUST_PC(pcAdjustment);
		    
		case TCL_BREAK:
		case TCL_CONTINUE:
		    /*
		     * The invoked command requested a break or continue.
		     * Find the closest enclosing loop or catch exception
		     * range, if any. If a loop is found, terminate its
		     * execution or skip to its next iteration. If the
		     * closest is a catch exception range, jump to its
		     * catchOffset. If no enclosing range is found, stop
		     * execution and return the TCL_BREAK or TCL_CONTINUE.
		     */
		    rangePtr = TclGetExceptionRangeForPc(pc,
                            /*catchOnly*/ 0, codePtr);
		    if (rangePtr == NULL) {
		        TRACE(("%s %u => ... after \"%.20s\", no encl. loop or catch, returning %s\n",
		                opName[opCode], objc, cmdNameBuf,
			        StringForResultCode(result)));
			goto abnormalReturn; /* no catch exists to check */
		    }
		    switch (rangePtr->type) {
		    case LOOP_EXCEPTION_RANGE:
			if (result == TCL_BREAK) {
			    newPcOffset = rangePtr->breakOffset;
			} else if (rangePtr->continueOffset == -1) {
			    TRACE(("%s %u => ... after \"%.20s\", %s, loop w/o continue, checking for catch\n",
				   opName[opCode], objc, cmdNameBuf,
				   StringForResultCode(result)));
			    goto checkForCatch;
			} else {
			    newPcOffset = rangePtr->continueOffset;
			}
			result = TCL_OK;
			TRACE(("%s %u => ... after \"%.20s\", %s, range at %d, new pc %d\n",
			       opName[opCode], objc, cmdNameBuf,
			       StringForResultCode(result),
			       rangePtr->codeOffset, newPcOffset));
			break;
		    case CATCH_EXCEPTION_RANGE:
			TRACE(("%s %u => ... after \"%.20s\", %s...\n",
			       opName[opCode], objc, cmdNameBuf,
			       StringForResultCode(result)));
			goto processCatch; /* it will use rangePtr */
		    default:
			panic("TclExecuteByteCode: unrecognized ExceptionRange type %d\n", rangePtr->type);
		    }
		    pc = (codePtr->codeStart + newPcOffset);
		    continue;	/* restart outer instruction loop at pc */
		    
		case TCL_ERROR:
		    /*
		     * The invoked command return an error. Record
		     * information about what was being executed when the
		     * error occurred, then look for an enclosing catch
		     * exception range, if any.
		     */
		    valuePtr = iPtr->objResult;
		    TRACE_WITH_OBJ(("%s %u => ... after \"%.20s\", TCL_ERROR ",
		            opName[opCode], objc, cmdNameBuf), valuePtr);
		    if (!(iPtr->flags & ERR_ALREADY_LOGGED)) {
                        char buf[200];
                        int cmdIndex = TclGetSrcInfoForPc(pc, codePtr);
                        if (cmdIndex != -1) {
                            CmdLocation *locPtr =
                                &(codePtr->cmdMapPtr[cmdIndex]);
                            char *ellipsis = "";
                            int numChars = locPtr->numSrcChars;
                            if (numChars > 150) {
                                numChars = 150;
                                ellipsis = "...";
                            }
                            if (!(iPtr->flags & ERR_IN_PROGRESS)) {
                                sprintf(buf, "\n    while executing\n\"%.*s%s\"",
                                        numChars,
                                        (codePtr->source + locPtr->srcOffset),
                                        ellipsis);
                            } else {
                                sprintf(buf, "\n    invoked from within\n\"%.*s%s\"",
                                        numChars, 
                                        (codePtr->source + locPtr->srcOffset),
                                        ellipsis);
                            }
                            Tcl_AddObjErrorInfo(interp, buf, -1);
                            iPtr->flags |= ERR_ALREADY_LOGGED;
                        }
                    }
		    goto checkForCatch;

		case TCL_RETURN:
		    /*
		     * The invoked command requested that the current
		     * procedure stop execution and return. First check
		     * for an enclosing catch exception range, if any.
		     */
		    TRACE(("%s %u => ... after \"%.20s\", TCL_RETURN\n",
		            opName[opCode], objc, cmdNameBuf));
		    goto checkForCatch;

		default:
		    TRACE_WITH_OBJ(("%s %u => ... after \"%.20s\", OTHER RETURN CODE %d ",
		            opName[opCode], objc, cmdNameBuf, result),
			    iPtr->objResult);
		    goto checkForCatch;
		} /* end of switch on result from invoke instruction */
	    }
	    
	case INST_EVAL_STK:
	    objPtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    result = Tcl_EvalObj(interp, objPtr);
	    CACHE_STACK_INFO();
	    if (result == TCL_OK) {
		/*
		 * Normal return; push the eval's object result.
		 */
		valuePtr = iPtr->objResult;
		Tcl_IncrRefCount(valuePtr); /* stack points to obj */
		stackPtr[++stackTop].o = valuePtr;
		TRACE_WITH_OBJ(("evalStk \"%.30s\" => ", O2S(objPtr)),
			valuePtr);
		TclDecrRefCount(objPtr);    /* done with popped object */
		ADJUST_PC(1);
	    } else if ((result == TCL_BREAK) || (result == TCL_CONTINUE)) {
		/*
		 * Find the closest enclosing loop or catch exception range,
		 * if any. If a loop is found, terminate its execution or
		 * skip to its next iteration. If the closest is a catch
		 * exception range, jump to its catchOffset. If no enclosing
		 * range is found, stop execution and return that same
		 * TCL_BREAK or TCL_CONTINUE.
		 */

		int newPcOffset = 0; /* Pc offset computed during break,
				      * continue, error processing. Init.
				      * to avoid compiler warning. */

		rangePtr = TclGetExceptionRangeForPc(pc, /*catchOnly*/ 0,
			codePtr);
		if (rangePtr == NULL) {
		    TRACE(("evalStk \"%.30s\" => no encl. loop or catch, returning %s\n",
			    O2S(objPtr), StringForResultCode(result)));
		    TclDecrRefCount(objPtr); /* done with popped obj */
		    goto abnormalReturn;    /* no catch exists to check */
		}
		switch (rangePtr->type) {
		case LOOP_EXCEPTION_RANGE:
		    if (result == TCL_BREAK) {
			newPcOffset = rangePtr->breakOffset;
		    } else if (rangePtr->continueOffset == -1) {
			TRACE(("evalStk \"%.30s\" => %s, loop w/o continue, checking for catch\n",
			       O2S(objPtr), StringForResultCode(result)));
			TclDecrRefCount(objPtr); /* done with popped obj */
			goto checkForCatch;
		    } else {
			newPcOffset = rangePtr->continueOffset;
		    }
		    result = TCL_OK;
		    TRACE_WITH_OBJ(("evalStk \"%.30s\" => %s, range at %d, new pc %d ",
			    O2S(objPtr), StringForResultCode(result),
			    rangePtr->codeOffset, newPcOffset), valuePtr);
		    break;
		case CATCH_EXCEPTION_RANGE:
		    TRACE_WITH_OBJ(("evalStk \"%.30s\" => %s ",
			    O2S(objPtr), StringForResultCode(result)),
			    valuePtr);
		    TclDecrRefCount(objPtr); /* done with popped obj */
		    goto processCatch;  /* it will use rangePtr */
		default:
		    panic("TclExecuteByteCode: unrecognized ExceptionRange type %d\n", rangePtr->type);
		}
		TclDecrRefCount(objPtr);	/* done with popped obj */
		pc = (codePtr->codeStart + newPcOffset);
		continue;	/* restart outer instruction loop at pc */
	    } else { /* eval returned TCL_ERROR, TCL_RETURN, unknown code */
		TRACE_WITH_OBJ(("evalStk \"%.30s\" => ERROR: ", O2S(objPtr)),
		        iPtr->objResult);
		TclDecrRefCount(objPtr);	 /* done with popped obj */
		goto checkForCatch;
	    }

	case INST_EXPR_STK:
	    objPtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    result = Tcl_ExprStringObj(interp, objPtr);
	    CACHE_STACK_INFO();
	    if (result != TCL_OK) {
		TRACE_WITH_OBJ(("exprStk \"%.30s\" => ERROR: ", 
		        O2S(objPtr)), iPtr->objResult);
		TclDecrRefCount(objPtr); /* done with popped object */
		goto checkForCatch;
	    }
	    valuePtr = iPtr->objResult;
	    Tcl_IncrRefCount(valuePtr);  /* stack points to obj */
	    stackPtr[++stackTop].o = valuePtr;
	    TRACE_WITH_OBJ(("exprStk \"%.30s\" => ", O2S(objPtr)),
		       valuePtr);
	    TclDecrRefCount(objPtr);     /* done with popped object */
	    ADJUST_PC(1);

	case INST_LOAD_SCALAR4:
	    opnd = TclGetInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doLoadScalar;

	case INST_LOAD_SCALAR1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doLoadScalar:
	    DECACHE_STACK_INFO();
	    valuePtr = TclGetIndexedScalar(interp, opnd,
					   /*leaveErrorMsg*/ 1);
	    CACHE_STACK_INFO();
	    if (valuePtr == NULL) {
		TRACE_WITH_OBJ(("%s %u => ERROR: ", opName[opCode], opnd),
			iPtr->objResult);
		result = TCL_ERROR;
		goto checkForCatch;
            }
	    PUSH_OBJECT(valuePtr);
	    TRACE_WITH_OBJ(("%s %u => ", opName[opCode], opnd), valuePtr);
	    ADJUST_PC(pcAdjustment);

	case INST_LOAD_SCALAR_STK:
	    namePtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    valuePtr = Tcl_ObjGetVar2(interp, namePtr, (Tcl_Obj *) NULL, 
				      TCL_LEAVE_ERR_MSG);
	    CACHE_STACK_INFO();
	    if (valuePtr == NULL) {
		TRACE_WITH_OBJ(("loadScalarStk \"%.30s\" => ERROR: ",
		        O2S(namePtr)), iPtr->objResult);
		TclDecrRefCount(namePtr); /* done with popped name. */
		result = TCL_ERROR;
		goto checkForCatch;
            }
	    PUSH_OBJECT(valuePtr);
	    TRACE_WITH_OBJ(("loadScalarStk \"%.30s\" => ",
		    O2S(namePtr)), valuePtr);
	    TclDecrRefCount(namePtr);     /* done with popped name. */
	    ADJUST_PC(1);

	case INST_LOAD_ARRAY4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doLoadArray;

	case INST_LOAD_ARRAY1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doLoadArray:
	    {
		Tcl_Obj *elemPtr = POP_OBJECT();
		
		DECACHE_STACK_INFO();
		valuePtr = TclGetElementOfIndexedArray(interp, opnd,
	                elemPtr, /*leaveErrorMsg*/ 1);
		CACHE_STACK_INFO();
		if (valuePtr == NULL) {
		    TRACE_WITH_OBJ(("%s %u \"%.30s\" => ERROR: ",
			    opName[opCode], opnd, O2S(elemPtr)),
			    iPtr->objResult);
		    TclDecrRefCount(elemPtr); /* done with element name. */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(valuePtr);
		TRACE_WITH_OBJ(("%s %u \"%.30s\" => ",
		        opName[opCode], opnd, O2S(elemPtr)), valuePtr);
		TclDecrRefCount(elemPtr);     /* done with element name. */
	    }
	    ADJUST_PC(pcAdjustment);

	case INST_LOAD_ARRAY_STK:
	    {
		Tcl_Obj *elemPtr = POP_OBJECT();
		
		namePtr = POP_OBJECT();
		DECACHE_STACK_INFO();
		valuePtr = Tcl_ObjGetVar2(interp, namePtr, elemPtr,
		        TCL_LEAVE_ERR_MSG);
		CACHE_STACK_INFO();
		if (valuePtr == NULL) {
		    TRACE_WITH_OBJ(("loadArrayStk \"%.30s(%.30s)\" => ERROR: ",
		            O2S(namePtr), O2S(elemPtr)), iPtr->objResult);
		    TclDecrRefCount(namePtr); /* done with array name */
		    TclDecrRefCount(elemPtr); /* and element name. */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(valuePtr);
		TRACE_WITH_OBJ(("loadArrayStk \"%.30s(%.30s)\" => ",
		        O2S(namePtr), O2S(elemPtr)), valuePtr);
		TclDecrRefCount(namePtr);     /* done with array name */
		TclDecrRefCount(elemPtr);     /* and element name. */
	    }
	    ADJUST_PC(1);

	case INST_LOAD_STK:
	    namePtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    valuePtr = Tcl_ObjGetVar2(interp, namePtr, NULL,
		    TCL_PART1_NOT_PARSED|TCL_LEAVE_ERR_MSG);
	    CACHE_STACK_INFO();
	    if (valuePtr == NULL) {
		TRACE_WITH_OBJ(("loadStk \"%.30s\" => ERROR: ",
		        O2S(namePtr)), iPtr->objResult);
		TclDecrRefCount(namePtr); /* done with popped name. */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(valuePtr);
	    TRACE_WITH_OBJ(("loadStk \"%.30s\" => ", O2S(namePtr)),
		    valuePtr);
	    TclDecrRefCount(namePtr);     /* done with popped name. */
	    ADJUST_PC(1);
	    
	case INST_STORE_SCALAR4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doStoreScalar;

	case INST_STORE_SCALAR1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	  doStoreScalar:
	    valuePtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    value2Ptr = TclSetIndexedScalar(interp, opnd, valuePtr,
					      /*leaveErrorMsg*/ 1);
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("%s %u <- \"%.30s\" => ERROR: ",
			opName[opCode], opnd, O2S(valuePtr)),
			iPtr->objResult);
		TclDecrRefCount(valuePtr); /* done with popped value. */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("%s %u <- \"%.30s\" => ",
		    opName[opCode], opnd, O2S(valuePtr)), value2Ptr);
	    TclDecrRefCount(valuePtr);     /* done with popped value. */
	    ADJUST_PC(pcAdjustment);

	case INST_STORE_SCALAR_STK:
	    valuePtr = POP_OBJECT();
	    namePtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    value2Ptr = Tcl_ObjSetVar2(interp, namePtr, NULL, valuePtr,
	            TCL_LEAVE_ERR_MSG);
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(
			("storeScalarStk \"%.30s\" <- \"%.30s\" => ERROR: ",
		        O2S(namePtr),
			O2S(valuePtr)),
			iPtr->objResult);
		TclDecrRefCount(namePtr);  /* done with popped name. */
		TclDecrRefCount(valuePtr); /* done with popped value. */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(
		    ("storeScalarStk \"%.30s\" <- \"%.30s\" => ",
		    O2S(namePtr),
		    O2S(valuePtr)),
		    value2Ptr);
	    TclDecrRefCount(namePtr);      /* done with popped name. */
	    TclDecrRefCount(valuePtr);     /* done with popped value. */
	    ADJUST_PC(1);

	case INST_STORE_ARRAY4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doStoreArray;

	case INST_STORE_ARRAY1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doStoreArray:
	    {
		Tcl_Obj *elemPtr;

		valuePtr = POP_OBJECT();
		elemPtr = POP_OBJECT();
		DECACHE_STACK_INFO();
		value2Ptr = TclSetElementOfIndexedArray(interp, opnd,
		        elemPtr, valuePtr, TCL_LEAVE_ERR_MSG);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(
			    ("%s %u \"%.30s\" <- \"%.30s\" => ERROR: ",
			    opName[opCode], opnd, O2S(elemPtr),
			    O2S(valuePtr)),
		            iPtr->objResult);
		    TclDecrRefCount(elemPtr);  /* done with element name. */
		    TclDecrRefCount(valuePtr); /* done with popped value. */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("%s %u \"%.30s\" <- \"%.30s\" => ",
		        opName[opCode], opnd, O2S(elemPtr), O2S(valuePtr)),
		        value2Ptr);
		TclDecrRefCount(elemPtr);      /* done with element name. */
		TclDecrRefCount(valuePtr);     /* done with popped value. */
	    }
	    ADJUST_PC(pcAdjustment);

	case INST_STORE_ARRAY_STK:
	    {
		Tcl_Obj *elemPtr;

		valuePtr = POP_OBJECT();
		elemPtr = POP_OBJECT();
		namePtr = POP_OBJECT();
		DECACHE_STACK_INFO();
		value2Ptr = Tcl_ObjSetVar2(interp, namePtr, elemPtr,
		        valuePtr, TCL_LEAVE_ERR_MSG);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(("storeArrayStk \"%.30s(%.30s)\" <- \"%.30s\" => ERROR: ",
		            O2S(namePtr), O2S(elemPtr), O2S(valuePtr)),
			    iPtr->objResult);
		    TclDecrRefCount(namePtr);  /* done with array name, */
		    TclDecrRefCount(elemPtr);  /* the element name, */
		    TclDecrRefCount(valuePtr); /* and the popped value. */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("storeArrayStk \"%.30s(%.30s)\" <- \"%.30s\" => ",
		        O2S(namePtr), O2S(elemPtr), O2S(valuePtr)),
			value2Ptr);
		TclDecrRefCount(namePtr);      /* done with array name, */
		TclDecrRefCount(elemPtr);      /* the element name, */
		TclDecrRefCount(valuePtr);     /* and popped value. */
	    }
	    ADJUST_PC(1);

	case INST_STORE_STK:
	    valuePtr = POP_OBJECT();
	    namePtr = POP_OBJECT();
	    DECACHE_STACK_INFO();
	    value2Ptr = Tcl_ObjSetVar2(interp, namePtr, NULL, valuePtr,
		    TCL_PART1_NOT_PARSED|TCL_LEAVE_ERR_MSG);
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("storeStk \"%.30s\" <- \"%.30s\" => ERROR: ",
		        O2S(namePtr), O2S(valuePtr)), iPtr->objResult);
		TclDecrRefCount(namePtr);  /* done with popped name. */
		TclDecrRefCount(valuePtr); /* and popped value. */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("storeStk \"%.30s\" <- \"%.30s\" => ",
		    O2S(namePtr), O2S(valuePtr)), value2Ptr);
	    TclDecrRefCount(namePtr);      /* done with popped name */
	    TclDecrRefCount(valuePtr);     /* and popped value. */
	    ADJUST_PC(1);

	case INST_INCR_SCALAR1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    valuePtr = POP_OBJECT(); 
	    if (valuePtr->typePtr != &tclIntType) {
		result = tclIntType.setFromAnyProc(interp, valuePtr);
		if (result != TCL_OK) {
		    TRACE_WITH_OBJ(("incrScalar1 %u (by %s) => ERROR converting increment amount to int: ",
		            opnd, O2S(valuePtr)), iPtr->objResult);
		    TclDecrRefCount(valuePtr);  /* done with incr amount */
		    goto checkForCatch;
		}
	    }
	    i = valuePtr->internalRep.intValue;
	    DECACHE_STACK_INFO();
	    value2Ptr = TclIncrIndexedScalar(interp, opnd, i);
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("incrScalar1 %u (by %ld) => ERROR: ",
		        opnd, i), iPtr->objResult);
		TclDecrRefCount(valuePtr);      /* done with incr amount */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("incrScalar1 %u (by %ld) => ", opnd, i),
		    value2Ptr);
	    TclDecrRefCount(valuePtr);          /* done with incr amount */
	    ADJUST_PC(2);

	case INST_INCR_SCALAR_STK:
	case INST_INCR_STK:
	    valuePtr = POP_OBJECT();
	    namePtr = POP_OBJECT();
	    if (valuePtr->typePtr != &tclIntType) {
		result = tclIntType.setFromAnyProc(interp, valuePtr);
		if (result != TCL_OK) {
		    TRACE_WITH_OBJ(("%s \"%.30s\" (by %s) => ERROR converting increment amount to int: ",
		            opName[opCode], O2S(namePtr), O2S(valuePtr)),
			    iPtr->objResult);
		    TclDecrRefCount(namePtr);   /* done with var name */
		    TclDecrRefCount(valuePtr);  /* done with incr amount */
		    goto checkForCatch;
		}
	    }
	    i = valuePtr->internalRep.intValue;
	    DECACHE_STACK_INFO();
	    value2Ptr = TclIncrVar2(interp, namePtr, (Tcl_Obj *) NULL, i,
	        /*part1NotParsed*/ (opCode == INST_INCR_STK));
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("%s \"%.30s\" (by %ld) => ERROR: ",
		        opName[opCode], O2S(namePtr), i), iPtr->objResult);
		TclDecrRefCount(namePtr);       /* done with var name */
		TclDecrRefCount(valuePtr);      /* done with incr amount */
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("%s \"%.30s\" (by %ld) => ",
	            opName[opCode], O2S(namePtr), i), value2Ptr);
	    TclDecrRefCount(namePtr);           /* done with var name */
	    TclDecrRefCount(valuePtr);          /* done with incr amount */
	    ADJUST_PC(1);

	case INST_INCR_ARRAY1:
	    {
		Tcl_Obj *elemPtr;

		opnd = TclGetUInt1AtPc(pc+1);
		valuePtr = POP_OBJECT();
		elemPtr = POP_OBJECT();
		if (valuePtr->typePtr != &tclIntType) {
		    result = tclIntType.setFromAnyProc(interp, valuePtr);
		    if (result != TCL_OK) {
			TRACE_WITH_OBJ(("incrArray1 %u \"%.30s\" (by %s) => ERROR converting increment amount to int: ",
		                opnd, O2S(elemPtr), O2S(valuePtr)),
			        iPtr->objResult);
			TclDecrRefCount(elemPtr);  /* done w/ element name */
			TclDecrRefCount(valuePtr); /* done w/ incr amount */
			goto checkForCatch;
		    }
		}
		i = valuePtr->internalRep.intValue;
		DECACHE_STACK_INFO();
		value2Ptr = TclIncrElementOfIndexedArray(interp, opnd,
		        elemPtr, i);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(("incrArray1 %u \"%.30s\" (by %ld) => ERROR: ",
		            opnd, O2S(elemPtr), i), iPtr->objResult);
		    TclDecrRefCount(elemPtr);   /* done with element name */
		    TclDecrRefCount(valuePtr);  /* done with incr amount */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("incrArray1 %u \"%.30s\" (by %ld) => ",
	                opnd, O2S(elemPtr), i), value2Ptr);
		TclDecrRefCount(elemPtr);       /* done with element name */
		TclDecrRefCount(valuePtr);      /* done with incr amount */
	    }
	    ADJUST_PC(2);
	    
	case INST_INCR_ARRAY_STK:
	    {
		Tcl_Obj *elemPtr;

		valuePtr = POP_OBJECT();
		elemPtr = POP_OBJECT();
		namePtr = POP_OBJECT();
		if (valuePtr->typePtr != &tclIntType) {
		    result = tclIntType.setFromAnyProc(interp, valuePtr);
		    if (result != TCL_OK) {
		        TRACE_WITH_OBJ(("incrArrayStk \"%.30s(%.30s)\" (by %s) => ERROR converting increment amount to int: ",
		                O2S(namePtr), O2S(elemPtr), O2S(valuePtr)),
			        iPtr->objResult);
			TclDecrRefCount(namePtr);  /* done w/ array name */
			TclDecrRefCount(elemPtr);  /* done w/ element name */
			TclDecrRefCount(valuePtr); /* done w/ incr amount */
			goto checkForCatch;
		    }
		}
		i = valuePtr->internalRep.intValue;
		DECACHE_STACK_INFO();
		value2Ptr = TclIncrVar2(interp, namePtr, elemPtr, i,
					/*part1NotParsed*/ 0);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(("incrArrayStk \"%.30s(%.30s)\" (by %ld) => ERROR: ",
		            O2S(namePtr), O2S(elemPtr), i),
			    iPtr->objResult);
		    TclDecrRefCount(namePtr);      /* done w/ array name */
		    TclDecrRefCount(elemPtr);      /* done w/ element name */
		    TclDecrRefCount(valuePtr);     /* done w/ incr amount */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("incrArrayStk \"%.30s(%.30s)\" (by %ld) => ",
	                O2S(namePtr), O2S(elemPtr), i), value2Ptr);
		TclDecrRefCount(namePtr);          /* done w/ array name */
		TclDecrRefCount(elemPtr);          /* done w/ element name */
		TclDecrRefCount(valuePtr);         /* done w/ incr amount */
	    }
	    ADJUST_PC(1);
	    
	case INST_INCR_SCALAR1_IMM:
	    opnd = TclGetUInt1AtPc(pc+1);
	    i = TclGetInt1AtPc(pc+2);
	    DECACHE_STACK_INFO();
	    value2Ptr = TclIncrIndexedScalar(interp, opnd, i);
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("incrScalar1Imm %u %ld => ERROR: ",
		        opnd, i), iPtr->objResult);
		result = TCL_ERROR;
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("incrScalar1Imm %u %ld => ", opnd, i),
		    value2Ptr);
	    ADJUST_PC(3);

	case INST_INCR_SCALAR_STK_IMM:
	case INST_INCR_STK_IMM:
	    namePtr = POP_OBJECT();
	    i = TclGetInt1AtPc(pc+1);
	    DECACHE_STACK_INFO();
	    value2Ptr = TclIncrVar2(interp, namePtr, (Tcl_Obj *) NULL, i,
		    /*part1NotParsed*/ (opCode == INST_INCR_STK_IMM));
	    CACHE_STACK_INFO();
	    if (value2Ptr == NULL) {
		TRACE_WITH_OBJ(("%s \"%.30s\" %ld => ERROR: ",
		        opName[opCode], O2S(namePtr), i), iPtr->objResult);
		result = TCL_ERROR;
		TclDecrRefCount(namePtr);  /* done with var name */
		goto checkForCatch;
	    }
	    PUSH_OBJECT(value2Ptr);
	    TRACE_WITH_OBJ(("%s \"%.30s\" %ld => ",
	            opName[opCode], O2S(namePtr), i), value2Ptr);
	    TclDecrRefCount(namePtr);  /* done with var name */
	    ADJUST_PC(2);

	case INST_INCR_ARRAY1_IMM:
	    {
		Tcl_Obj *elemPtr;

		opnd = TclGetUInt1AtPc(pc+1);
		i = TclGetInt1AtPc(pc+2);
		elemPtr = POP_OBJECT();
		DECACHE_STACK_INFO();
		value2Ptr = TclIncrElementOfIndexedArray(interp, opnd,
		        elemPtr, i);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(("incrArray1Imm %u \"%.30s\" (by %ld) => ERROR: ",
		            opnd, O2S(elemPtr), i), iPtr->objResult);
		    TclDecrRefCount(elemPtr); /* done with element name */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("incrArray1Imm %u \"%.30s\" (by %ld) => ",
	                opnd, O2S(elemPtr), i), value2Ptr);
		TclDecrRefCount(elemPtr);     /* done with element name */
	    }
	    ADJUST_PC(3);
	    
	case INST_INCR_ARRAY_STK_IMM:
	    {
		Tcl_Obj *elemPtr;

		i = TclGetInt1AtPc(pc+1);
		elemPtr = POP_OBJECT();
		namePtr = POP_OBJECT();
		DECACHE_STACK_INFO();
		value2Ptr = TclIncrVar2(interp, namePtr, elemPtr, i,
		        /*part1NotParsed*/ 0);
		CACHE_STACK_INFO();
		if (value2Ptr == NULL) {
		    TRACE_WITH_OBJ(("incrArrayStkImm \"%.30s(%.30s)\" (by %ld) => ERROR: ",
		            O2S(namePtr), O2S(elemPtr), i),
			    iPtr->objResult);
		    TclDecrRefCount(namePtr); /* done with array name */
		    TclDecrRefCount(elemPtr); /* done with element name */
		    result = TCL_ERROR;
		    goto checkForCatch;
		}
		PUSH_OBJECT(value2Ptr);
		TRACE_WITH_OBJ(("incrArrayStkImm \"%.30s(%.30s)\" (by %ld) => ",
	                O2S(namePtr), O2S(elemPtr), i), value2Ptr);
		TclDecrRefCount(namePtr);     /* done with array name */
		TclDecrRefCount(elemPtr);     /* done with element name */
	    }
	    ADJUST_PC(2);

	case INST_JUMP1:
	    opnd = TclGetInt1AtPc(pc+1);
	    TRACE(("jump1 %d => new pc %u\n", opnd,
		   (unsigned int)(pc + opnd - codePtr->codeStart)));
	    ADJUST_PC(opnd);

	case INST_JUMP4:
	    opnd = TclGetInt4AtPc(pc+1);
	    TRACE(("jump4 %d => new pc %u\n", opnd,
		   (unsigned int)(pc + opnd - codePtr->codeStart)));
	    ADJUST_PC(opnd);

	case INST_JUMP_TRUE4:
	    opnd = TclGetInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doJumpTrue;

	case INST_JUMP_TRUE1:
	    opnd = TclGetInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doJumpTrue:
	    {
		int b;
		
		valuePtr = POP_OBJECT();
		if (valuePtr->typePtr == &tclIntType) {
		    b = (valuePtr->internalRep.intValue != 0);
		} else if (valuePtr->typePtr == &tclDoubleType) {
		    b = (valuePtr->internalRep.doubleValue != 0.0);
		} else {
		    result = Tcl_GetBooleanFromObj(interp, valuePtr, &b);
		    if (result != TCL_OK) {
			TRACE_WITH_OBJ(("%s %d => ERROR: ", opName[opCode],
				opnd), iPtr->objResult);
			TclDecrRefCount(valuePtr); /* done with popped obj */
			goto checkForCatch;
		    }
		}
		if (b) {
		    TRACE(("%s %d => %.20s true, new pc %u\n",
			    opName[opCode], opnd, O2S(valuePtr),
		            (unsigned int)(pc+opnd - codePtr->codeStart)));
		    TclDecrRefCount(valuePtr); /* done with popped obj */
		    ADJUST_PC(opnd);
		} else {
		    TRACE(("%s %d => %.20s false\n", opName[opCode], opnd,
		            O2S(valuePtr)));
		    TclDecrRefCount(valuePtr); /* done with popped obj */
		    ADJUST_PC(pcAdjustment);
		}
	    }
	    
	case INST_JUMP_FALSE4:
	    opnd = TclGetInt4AtPc(pc+1);
	    pcAdjustment = 5;
	    goto doJumpFalse;

	case INST_JUMP_FALSE1:
	    opnd = TclGetInt1AtPc(pc+1);
	    pcAdjustment = 2;
	    
	    doJumpFalse:
	    {
		int b;
		
		valuePtr = POP_OBJECT();
		if (valuePtr->typePtr == &tclIntType) {
		    b = (valuePtr->internalRep.intValue != 0);
		} else if (valuePtr->typePtr == &tclDoubleType) {
		    b = (valuePtr->internalRep.doubleValue != 0.0);
		} else {
		    result = Tcl_GetBooleanFromObj(interp, valuePtr, &b);
		    if (result != TCL_OK) {
			TRACE_WITH_OBJ(("%s %d => ERROR: ", opName[opCode],
				opnd), iPtr->objResult);
			TclDecrRefCount(valuePtr); /* done with popped obj */
			goto checkForCatch;
		    }
		}
		if (b) {
		    TRACE(("%s %d => %.20s true\n", opName[opCode], opnd,
		            O2S(valuePtr)));
		    TclDecrRefCount(valuePtr); /* done with popped obj */
		    ADJUST_PC(pcAdjustment);
		} else {
		    TRACE(("%s %d => %.20s false, new pc %u\n",
			    opName[opCode], opnd, O2S(valuePtr),
			   (unsigned int)(pc + opnd - codePtr->codeStart)));
		    TclDecrRefCount(valuePtr); /* done with popped obj */
		    ADJUST_PC(opnd);
		}
	    }
	    
	case INST_LOR:
	case INST_LAND:
	    {
		/*
		 * Operands must be numeric, but no int->double conversions
		 * are performed.
		 */
		
		long i2, iResult;
		double d1;
		char *s;
		
		value2Ptr = POP_OBJECT();
		valuePtr  = POP_OBJECT();
		
		if (valuePtr->typePtr == &tclIntType) {
		    i = (valuePtr->internalRep.intValue != 0);
		} else if (valuePtr->typePtr == &tclDoubleType) {
		    i = (valuePtr->internalRep.doubleValue != 0.0);
		} else {
		    s = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s)) {
			result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, valuePtr, &d1);
		    }
		    if (result == TCL_OK) {
			if (valuePtr->typePtr == &tclIntType) {
			    i = (valuePtr->internalRep.intValue != 0);
			} else if (valuePtr->typePtr == &tclDoubleType) {
			    i = (valuePtr->internalRep.doubleValue != 0.0);
			}
		    } else {
			TRACE(("%s \"%.20s\" => ILLEGAL TYPE %s \n",
			        opName[opCode], O2S(valuePtr),
			        (valuePtr->typePtr?
				    valuePtr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, valuePtr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		}
		
		if (value2Ptr->typePtr == &tclIntType) {
		    i2 = (value2Ptr->internalRep.intValue != 0);
		} else if (value2Ptr->typePtr == &tclDoubleType) {
		    i2 = (value2Ptr->internalRep.doubleValue != 0.0);
		} else {
		    s = TclGetStringFromObj(value2Ptr, &length);
		    if (TclLooksLikeInt(s)) {
			result = Tcl_GetIntFromObj(interp, value2Ptr, &i2);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, value2Ptr, &d1);
		    }
		    if (result == TCL_OK) {
			if (value2Ptr->typePtr == &tclIntType) {
			    i2 = (value2Ptr->internalRep.intValue != 0);
			} else if (valuePtr->typePtr == &tclDoubleType) {
			    i2 = (value2Ptr->internalRep.doubleValue != 0.0);
			}
		    } else {
			TRACE(("%s \"%.20s\" => ILLEGAL TYPE %s \n",
			        opName[opCode], O2S(value2Ptr),
			        (value2Ptr->typePtr?
				    value2Ptr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, value2Ptr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		}
		
		if (opCode == INST_LOR) {
		    iResult = (i || i2);
		} else {
		    iResult = (i && i2);
		}

		/*
		 * Reuse the valuePtr object already on stack if possible.
		 */
		
		if (Tcl_IsShared(valuePtr)) {
		    TclNewObj(objPtr);
		    objPtr->internalRep.intValue = iResult;
		    objPtr->typePtr = &tclIntType;
		    stackPtr[++stackTop].o = objPtr;
		    TRACE(("%s %.20s %.20s => %ld\n", opName[opCode],
			   O2S(valuePtr), O2S(value2Ptr), iResult));
		    TclDecrRefCount(valuePtr);    /* done with object */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		} else {	/* reuse the valuePtr object */
		    TRACE(("%s %.20s %.20s => %ld\n", 
			   opName[opCode], /* NB: stack top is off by 1 */
			   O2S(valuePtr), O2S(value2Ptr), iResult));
		    valuePtr->internalRep.intValue = iResult;
		    valuePtr->typePtr = &tclIntType;
		    if (valuePtr->bytes != NULL) {
			ckfree((char *) valuePtr->bytes);
			valuePtr->bytes = NULL;
		    }
		    ++stackTop; /* valuePtr now on stk top has right r.c. */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		}
	    }
	    ADJUST_PC(1);

	case INST_EQ:
	case INST_NEQ:
	case INST_LT:
	case INST_GT:
	case INST_LE:
	case INST_GE:
	    {
		/*
		 * Any type is allowed but the two operands must have the
	         * same type. We will compute value op value2.
		 */

		Tcl_ObjType *t1Ptr, *t2Ptr;
		char *s1 = NULL;   /* Init. avoids compiler warning. */
		char *s2 = NULL;   /* Init. avoids compiler warning. */
		long i2 = 0;	   /* Init. avoids compiler warning. */
		double d1 = 0.0;   /* Init. avoids compiler warning. */
		double d2 = 0.0;   /* Init. avoids compiler warning. */
		long iResult = 0;  /* Init. avoids compiler warning. */
		
		value2Ptr = POP_OBJECT();
		valuePtr  = POP_OBJECT();
		t1Ptr = valuePtr->typePtr;
		t2Ptr = value2Ptr->typePtr;
		if ((t1Ptr != &tclIntType) && (t1Ptr != &tclDoubleType)) {
		    s1 = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s1)) { /* FAILS IF NULLS */
			(void) Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			(void) Tcl_GetDoubleFromObj(interp, valuePtr, &d1);
		    }
		    t1Ptr = valuePtr->typePtr;
		}
		if ((t2Ptr != &tclIntType) && (t2Ptr != &tclDoubleType)) {
		    s2 = TclGetStringFromObj(value2Ptr, &length);
		    if (TclLooksLikeInt(s2)) { /* FAILS IF NULLS */
			(void) Tcl_GetIntFromObj(interp, value2Ptr, &i2);
		    } else {
			(void) Tcl_GetDoubleFromObj(interp, value2Ptr, &d2);
		    }
		    t2Ptr = value2Ptr->typePtr;
		}

		if (((t1Ptr != &tclIntType) && (t1Ptr != &tclDoubleType))
		        || ((t2Ptr != &tclIntType) && (t2Ptr != &tclDoubleType))) {
		    /*
		     * One operand is not numeric. Compare as strings.
		     * THIS FAILS IF AN OBJECT'S STRING REP CONTAINS NULLS.
		     */
		    int cmpValue;
		    s1 = TclGetStringFromObj(valuePtr, &length);
		    s2 = TclGetStringFromObj(value2Ptr, &length);
		    cmpValue = strcmp(s1, s2);
		    switch (opCode) {
		    case INST_EQ:
			iResult = (cmpValue == 0);
			break;
		    case INST_NEQ:
			iResult = (cmpValue != 0);
			break;
		    case INST_LT:
			iResult = (cmpValue < 0);
			break;
		    case INST_GT:
			iResult = (cmpValue > 0);
			break;
		    case INST_LE:
			iResult = (cmpValue <= 0);
			break;
		    case INST_GE:
			iResult = (cmpValue >= 0);
			break;
		    }
		} else if ((t1Ptr == &tclDoubleType)
		        || (t2Ptr == &tclDoubleType)) {
		    /*
		     * Compare as doubles.
		     */
		    if (t1Ptr == &tclDoubleType) {
			d1 = valuePtr->internalRep.doubleValue;
			if (t2Ptr == &tclIntType) {
			    d2 = value2Ptr->internalRep.intValue;
			} else {
			    d2 = value2Ptr->internalRep.doubleValue;
			}
		    } else {	/* t1Ptr is int, t2Ptr is double */
			d1 = valuePtr->internalRep.intValue;
			d2 = value2Ptr->internalRep.doubleValue;
		    }
		    switch (opCode) {
		    case INST_EQ:
			iResult = d1 == d2;
			break;
		    case INST_NEQ:
			iResult = d1 != d2;
			break;
		    case INST_LT:
			iResult = d1 < d2;
			break;
		    case INST_GT:
			iResult = d1 > d2;
			break;
		    case INST_LE:
			iResult = d1 <= d2;
			break;
		    case INST_GE:
			iResult = d1 >= d2;
			break;
		    }
		} else {
		    /*
		     * Compare as ints.
		     */
		    i  = valuePtr->internalRep.intValue;
		    i2 = value2Ptr->internalRep.intValue;
		    switch (opCode) {
		    case INST_EQ:
			iResult = i == i2;
			break;
		    case INST_NEQ:
			iResult = i != i2;
			break;
		    case INST_LT:
			iResult = i < i2;
			break;
		    case INST_GT:
			iResult = i > i2;
			break;
		    case INST_LE:
			iResult = i <= i2;
			break;
		    case INST_GE:
			iResult = i >= i2;
			break;
		    }
		}

		/*
		 * Reuse the valuePtr object already on stack if possible.
		 */
		
		if (Tcl_IsShared(valuePtr)) {
		    TclNewObj(objPtr);
		    objPtr->internalRep.intValue = iResult;
		    objPtr->typePtr = &tclIntType;
		    stackPtr[++stackTop].o = objPtr;
		    TRACE(("%s %.20s %.20s => %ld\n", opName[opCode],
		        O2S(valuePtr), O2S(value2Ptr), iResult));
		    TclDecrRefCount(valuePtr);    /* done with object */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		} else {	/* reuse the valuePtr object */
		    TRACE(("%s %.20s %.20s => %ld\n",
			opName[opCode], /* NB: stack top is off by 1 */
		        O2S(valuePtr), O2S(value2Ptr), iResult));
		    valuePtr->internalRep.intValue = iResult;
		    valuePtr->typePtr = &tclIntType;
		    if (valuePtr->bytes != NULL) {
			ckfree((char *) valuePtr->bytes);
			valuePtr->bytes = NULL;
		    }
		    ++stackTop; /* valuePtr now on stk top has right r.c. */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		}
	    }
	    ADJUST_PC(1);
	    
	case INST_MOD:
	case INST_LSHIFT:
	case INST_RSHIFT:
	case INST_BITOR:
	case INST_BITXOR:
	case INST_BITAND:
	    {
		/*
		 * Only integers are allowed. We compute value op value2.
		 */

		long i2, rem, negative;
		long iResult = 0; /* Init. avoids compiler warning. */
		
		value2Ptr = POP_OBJECT();
		valuePtr  = POP_OBJECT(); 
		if (valuePtr->typePtr == &tclIntType) {
		    i = valuePtr->internalRep.intValue;
		} else {	/* try to convert to int */
		    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    if (result != TCL_OK) {
			TRACE(("%s %.20s %.20s => ILLEGAL 1st TYPE %s\n",
			      opName[opCode], O2S(valuePtr), O2S(value2Ptr),
			      (valuePtr->typePtr?
				   valuePtr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, valuePtr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		}
		if (value2Ptr->typePtr == &tclIntType) {
		    i2 = value2Ptr->internalRep.intValue;
		} else {
		    result = Tcl_GetIntFromObj(interp, value2Ptr, &i2);
		    if (result != TCL_OK) {
			TRACE(("%s %.20s %.20s => ILLEGAL 2nd TYPE %s\n",
			      opName[opCode], O2S(valuePtr), O2S(value2Ptr),
			      (value2Ptr->typePtr?
				   value2Ptr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, value2Ptr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		}

		switch (opCode) {
		case INST_MOD:
		    /*
		     * This code is tricky: C doesn't guarantee much about
		     * the quotient or remainder, but Tcl does. The
		     * remainder always has the same sign as the divisor and
		     * a smaller absolute value.
		     */
		    if (i2 == 0) {
			TRACE(("mod %ld %ld => DIVIDE BY ZERO\n", i, i2));
			TclDecrRefCount(valuePtr);   /* done with object */
			TclDecrRefCount(value2Ptr);  /* done with object */
			goto divideByZero;
		    }
		    negative = 0;
		    if (i2 < 0) {
			i2 = -i2;
			i = -i;
			negative = 1;
		    }
		    rem  = i % i2;
		    if (rem < 0) {
			rem += i2;
		    }
		    if (negative) {
			rem = -rem;
		    }
		    iResult = rem;
		    break;
		case INST_LSHIFT:
		    iResult = i << i2;
		    break;
		case INST_RSHIFT:
		    /*
		     * The following code is a bit tricky: it ensures that
		     * right shifts propagate the sign bit even on machines
		     * where ">>" won't do it by default.
		     */
		    if (i < 0) {
			iResult = ~((~i) >> i2);
		    } else {
			iResult = i >> i2;
		    }
		    break;
		case INST_BITOR:
		    iResult = i | i2;
		    break;
		case INST_BITXOR:
		    iResult = i ^ i2;
		    break;
		case INST_BITAND:
		    iResult = i & i2;
		    break;
		}

		/*
		 * Reuse the valuePtr object already on stack if possible.
		 */
		
		if (Tcl_IsShared(valuePtr)) {
		    TclNewObj(objPtr);
		    objPtr->internalRep.intValue = iResult;
		    objPtr->typePtr = &tclIntType;
		    stackPtr[++stackTop].o = objPtr;
		    TRACE(("%s %ld %ld => %ld\n", opName[opCode], i, i2,
			   iResult));
		    TclDecrRefCount(valuePtr);    /* done with object */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		} else {	/* reuse the valuePtr object */
		    TRACE(("%s %ld %ld => %ld\n", opName[opCode], i, i2,
		        iResult)); /* NB: stack top is off by 1 */
		    valuePtr->internalRep.intValue = iResult;
		    valuePtr->typePtr = &tclIntType;
		    if (valuePtr->bytes != NULL) {
			ckfree((char *) valuePtr->bytes);
			valuePtr->bytes = NULL;
		    }
		    ++stackTop; /* valuePtr now on stk top has right r.c. */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		}
	    }
	    ADJUST_PC(1);
	    
	case INST_ADD:
	case INST_SUB:
	case INST_MULT:
	case INST_DIV:
	    {
		/*
		 * Operands must be numeric and ints get converted to floats
		 * if necessary. We compute value op value2.
		 */

		Tcl_ObjType *t1Ptr, *t2Ptr;
		long i2, quot, rem;
		double d1, d2;
		long iResult = 0;     /* Init. avoids compiler warning. */
		double dResult = 0.0; /* Init. avoids compiler warning. */
		int doDouble = 0;     /* 1 if doing floating arithmetic */
		
		value2Ptr = POP_OBJECT();
		valuePtr  = POP_OBJECT();
		t1Ptr = valuePtr->typePtr;
		t2Ptr = value2Ptr->typePtr;
		if (t1Ptr == &tclIntType) {
		    i  = valuePtr->internalRep.intValue;
		} else if (t1Ptr == &tclDoubleType) {
		    d1 = valuePtr->internalRep.doubleValue;
		} else {	     /* try to convert; FAILS IF NULLS */
		    char *s = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s)) {
			result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, valuePtr, &d1);
		    }
		    if (result != TCL_OK) {
			TRACE(("%s %.20s %.20s => ILLEGAL 1st TYPE %s\n",
			       opName[opCode], s, O2S(value2Ptr),
			       (valuePtr->typePtr?
				    valuePtr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, valuePtr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		    t1Ptr = valuePtr->typePtr;
		}
		if (t2Ptr == &tclIntType) {
		    i2 = value2Ptr->internalRep.intValue;
		} else if (t2Ptr == &tclDoubleType) {
		    d2 = value2Ptr->internalRep.doubleValue;
		} else {	     /* try to convert; FAILS IF NULLS */
		    char *s = TclGetStringFromObj(value2Ptr, &length);
		    if (TclLooksLikeInt(s)) {
			result = Tcl_GetIntFromObj(interp, value2Ptr, &i2);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, value2Ptr, &d2);
		    }
		    if (result != TCL_OK) {
			TRACE(("%s %.20s %.20s => ILLEGAL 2nd TYPE %s\n",
			       opName[opCode], O2S(valuePtr), s,
			       (value2Ptr->typePtr?
				    value2Ptr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, value2Ptr);
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		    t2Ptr = value2Ptr->typePtr;
		}

		if ((t1Ptr == &tclDoubleType) || (t2Ptr == &tclDoubleType)) {
		    /*
		     * Do double arithmetic.
		     */
		    doDouble = 1;
		    if (t1Ptr == &tclIntType) {
			d1 = i;       /* promote value 1 to double */
		    } else if (t2Ptr == &tclIntType) {
			d2 = i2;      /* promote value 2 to double */
		    }
		    switch (opCode) {
		    case INST_ADD:
			dResult = d1 + d2;
			break;
		    case INST_SUB:
			dResult = d1 - d2;
			break;
		    case INST_MULT:
			dResult = d1 * d2;
			break;
		    case INST_DIV:
			if (d2 == 0.0) {
			    TRACE(("div %.6g %.6g => DIVIDE BY ZERO\n",
				   d1, d2));
			    TclDecrRefCount(valuePtr);  /* done with obj */
			    TclDecrRefCount(value2Ptr); /* done with obj */
			    goto divideByZero;
			}
			dResult = d1 / d2;
			break;
		    }
		    
		    /*
		     * Check now for IEEE floating-point error.
		     */
		    
		    if (IS_NAN(dResult) || IS_INF(dResult)) {
			TRACE(("%s %.20s %.20s => IEEE FLOATING PT ERROR\n",
			       opName[opCode], O2S(valuePtr), O2S(value2Ptr)));
			TclExprFloatError(interp, dResult);
			result = TCL_ERROR;
			TclDecrRefCount(valuePtr);  /* done with object */
			TclDecrRefCount(value2Ptr); /* done with object */
			goto checkForCatch;
		    }
		} else {
		    /*
		     * Do integer arithmetic.
		     */
		    switch (opCode) {
		    case INST_ADD:
			iResult = i + i2;
			break;
		    case INST_SUB:
			iResult = i - i2;
			break;
		    case INST_MULT:
			iResult = i * i2;
			break;
		    case INST_DIV:
			/*
			 * This code is tricky: C doesn't guarantee much
			 * about the quotient or remainder, but Tcl does.
			 * The remainder always has the same sign as the
			 * divisor and a smaller absolute value.
			 */
			if (i2 == 0) {
			    TRACE(("div %ld %ld => DIVIDE BY ZERO\n",
				    i, i2));
			    TclDecrRefCount(valuePtr);  /* done with obj */
			    TclDecrRefCount(value2Ptr); /* done with obj */
			    goto divideByZero;
			}
			if (i2 < 0) {
			    i2 = -i2;
			    i = -i;
			}
			quot = i / i2;
			rem  = i % i2;
			if (rem < 0) {
			    quot -= 1;
			}
			iResult = quot;
			break;
		    }
		}

		/*
		 * Reuse the valuePtr object already on stack if possible.
		 */
		
		if (Tcl_IsShared(valuePtr)) {
		    TclNewObj(objPtr);
		    if (doDouble) {
			objPtr->internalRep.doubleValue = dResult;
			objPtr->typePtr = &tclDoubleType;
			stackPtr[++stackTop].o = objPtr;
			TRACE(("%s %.6g %.6g => %.6g\n", opName[opCode],
			       d1, d2, dResult));
		    } else {
			objPtr->internalRep.intValue = iResult;
			objPtr->typePtr = &tclIntType;
			stackPtr[++stackTop].o = objPtr;
			TRACE(("%s %ld %ld => %ld\n", opName[opCode],
			       i, i2, iResult));
		    } 
		    TclDecrRefCount(valuePtr);    /* done with object */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		} else {	    /* reuse the valuePtr object */
		    if (doDouble) { /* NB: stack top is off by 1 */
			TRACE(("%s %.6g %.6g => %.6g\n", opName[opCode],
			       d1, d2, dResult));
			valuePtr->internalRep.doubleValue = dResult;
			valuePtr->typePtr = &tclDoubleType;
		    } else {
			TRACE(("%s %ld %ld => %ld\n", opName[opCode],
			       i, i2, iResult));
			valuePtr->internalRep.intValue = iResult;
			valuePtr->typePtr = &tclIntType;
		    }
		    if (valuePtr->bytes != NULL) {
			ckfree((char *) valuePtr->bytes);
			valuePtr->bytes = NULL;
		    }
		    ++stackTop; /* valuePtr now on stk top has right r.c. */
		    TclDecrRefCount(value2Ptr);   /* done with object */
		}
	    }
	    ADJUST_PC(1);
	    
	case INST_UPLUS:
	    {
	        /*
	         * Operand must be numeric.
	         */

		double d;
		
		valuePtr = stackPtr[stackTop].o;
		if ((valuePtr->typePtr != &tclIntType)
		        && (valuePtr->typePtr != &tclDoubleType)) {
		    char *s = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s)) { /* FAILS IF NULLS */
			result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
		    }
		    if (result != TCL_OK) { 
			TRACE(("%s \"%.20s\" => ILLEGAL TYPE %s \n",
			        opName[opCode], s,
			        (valuePtr->typePtr?
			            valuePtr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, valuePtr);
			goto checkForCatch;
		    }
		}
		TRACE_WITH_OBJ(("uplus %s => ", O2S(valuePtr)), valuePtr);
	    }
	    ADJUST_PC(1);
	    
	case INST_UMINUS:
	case INST_LNOT:
	    {
		/*
		 * The operand must be numeric. If the operand object is
		 * unshared modify it directly, otherwise create a copy to
		 * modify: this is "copy on write". free any old string
		 * representation since it is now invalid.
		 */
		
		double d;
		
		valuePtr = POP_OBJECT();
		if ((valuePtr->typePtr != &tclIntType)
		        && (valuePtr->typePtr != &tclDoubleType)) {
		    char *s = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s)) { /* FAILS IF NULLS */
			result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
		    }
		    if (result != TCL_OK) {
			TRACE(("%s \"%.20s\" => ILLEGAL TYPE %s\n",
			        opName[opCode], s,
			        (valuePtr->typePtr?
				    valuePtr->typePtr->name : "null")));
			IllegalExprOperandType(interp, opCode, valuePtr);
			TclDecrRefCount(valuePtr);  /* done with object */
			goto checkForCatch;
		    }
		}
		
		if (Tcl_IsShared(valuePtr)) {
		    /*
		     * Create a new object.
		     */
		    TclNewObj(objPtr);
		    if (valuePtr->typePtr == &tclIntType) {
			i = valuePtr->internalRep.intValue;
			if (opCode == INST_UMINUS) {
			    objPtr->internalRep.intValue = -i;
			} else {
			    objPtr->internalRep.intValue = !i;
			}
			objPtr->typePtr = &tclIntType;
			TRACE_WITH_OBJ(("%s %ld => ", opName[opCode], i),
		                objPtr); /* NB: stack top is off by 1 */
		    } else {
			d = valuePtr->internalRep.doubleValue;
			if (opCode == INST_UMINUS) {
			    objPtr->internalRep.doubleValue = -d;
			    objPtr->typePtr = &tclDoubleType;
			} else {
			    /*
			     * Should be able to use "!d", but apparently
			     * some compilers can't handle it.
			     */
			    objPtr->internalRep.intValue = ((d==0.0)? 1:0);
			    objPtr->typePtr = &tclIntType;
			}
			TRACE_WITH_OBJ(("%s %.6g => ", opName[opCode], d),
		                objPtr); /* NB: stack top is off by 1 */
		    }
		    TclDecrRefCount(valuePtr); /* done with popped obj */
		    stackPtr[++stackTop].o = objPtr;
		} else {
		    /*
		     * valuePtr is unshared. Modify it directly.
		     */
		    if (valuePtr->bytes != NULL) {
			ckfree((char *) valuePtr->bytes);
			valuePtr->bytes = NULL;
		    }
		    if (valuePtr->typePtr == &tclIntType) {
			i = valuePtr->internalRep.intValue;
			if (opCode == INST_UMINUS) {
			    valuePtr->internalRep.intValue = -i;
			} else {
			    valuePtr->internalRep.intValue = !i;
			}
			TRACE_WITH_OBJ(("%s %ld => ", opName[opCode], i),
		                valuePtr); /* NB: stack top is off by 1 */
		    } else {
			d = valuePtr->internalRep.doubleValue;
			if (opCode == INST_UMINUS) {
			    valuePtr->internalRep.doubleValue = -d;
			} else {
			    /*
			     * Should be able to use "!d", but apparently
			     * some compilers can't handle it.
			     */
			    valuePtr->internalRep.intValue = ((d==0.0)?1:0);
			    valuePtr->typePtr = &tclIntType;
			}
			TRACE_WITH_OBJ(("%s %.6g => ", opName[opCode], d),
		                valuePtr); /* NB: stack top is off by 1 */
		    }
		    ++stackTop; /* valuePtr now on stk top has right r.c. */
		}
	    }
	    ADJUST_PC(1);
	    
	case INST_BITNOT:
	    /*
	     * The operand must be an integer. If the operand object is
	     * unshared modify it directly, otherwise modify a copy. 
	     * Free any old string representation since it is now invalid.
	     */
	
	    valuePtr = POP_OBJECT();
	    if (valuePtr->typePtr != &tclIntType) {
		result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		if (result != TCL_OK) {   /* try to convert to double */
		    TRACE(("bitnot \"%.20s\" => ILLEGAL TYPE %s\n",
		            O2S(valuePtr),
			    (valuePtr->typePtr?
			        valuePtr->typePtr->name : "null")));
		    IllegalExprOperandType(interp, opCode, valuePtr);
		    TclDecrRefCount(valuePtr);  /* done with object */
		    goto checkForCatch;
		}
	    }
	    if (Tcl_IsShared(valuePtr)) {
		/*
		 * Create a new object.
		 */
		i = valuePtr->internalRep.intValue;
		TclDecrRefCount(valuePtr);  /* done with popped obj */

		TclNewObj(objPtr);
		objPtr->internalRep.intValue = ~i;
		objPtr->typePtr = &tclIntType;
		stackPtr[++stackTop].o = objPtr; /* ref count is correct */
		TRACE_WITH_OBJ(("bitnot 0x%lx => (%lu) ", i, ~i), objPtr);
	    } else {
		/*
		 * valuePtr is unshared. Modify it directly.
		 */
		i = valuePtr->internalRep.intValue;
		valuePtr->internalRep.intValue = ~i;
		if (valuePtr->bytes != NULL) {
		    ckfree((char *) valuePtr->bytes);
		    valuePtr->bytes = NULL;
		}
		++stackTop; /* valuePtr now on stk top has right r.c. */
		TRACE_WITH_OBJ(("bitnot 0x%lx => (%lu) ", i, ~i), valuePtr);
	    }
	    ADJUST_PC(1);
	    
	case INST_CALL_BUILTIN_FUNC1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    {
		/*
		 * Call one of the built-in Tcl math functions.
		 */

		BuiltinFunc *mathFuncPtr;

		if ((opnd < 0) || (opnd > LAST_BUILTIN_FUNC)) {
		    TRACE(("UNRECOGNIZED BUILTIN FUNC CODE %d\n", opnd));
		    panic("TclExecuteByteCode: unrecognized builtin function code %d", opnd);
		}
		mathFuncPtr = &(builtinFuncTable[opnd]);
		DECACHE_STACK_INFO();
		tcl_MathInProgress++;
		result = (*mathFuncPtr->proc)(interp, eePtr,
		        mathFuncPtr->clientData);
		tcl_MathInProgress--;
		CACHE_STACK_INFO();
		if (result != TCL_OK) {
		    goto checkForCatch;
		}
		TRACE_WITH_OBJ(("callBuiltinFunc1 %d => ", opnd),
		        stackPtr[stackTop].o);
	    }
	    ADJUST_PC(2);
		    
	case INST_CALL_FUNC1:
	    opnd = TclGetUInt1AtPc(pc+1);
	    {
		/*
		 * Call a non-builtin Tcl math function previously
		 * registered by a call to Tcl_CreateMathFunc.
		 */
		
		int objc = opnd;   /* Number of arguments. The function name
				    * is the 0-th argument. */
		Tcl_Obj **objv;	   /* The array of arguments. The function
				    * name is objv[0]. */
		
		objv = &(stackPtr[stackTop - (objc-1)].o); /* "objv[0]" */
		DECACHE_STACK_INFO();
		tcl_MathInProgress++;
		result = ExprCallMathFunc(interp, eePtr, objc, objv);
		tcl_MathInProgress--;
		CACHE_STACK_INFO();
		if (result != TCL_OK) {
		    goto checkForCatch;
		}
		TRACE_WITH_OBJ(("callFunc1 %d => ", objc),
		        stackPtr[stackTop].o);
		ADJUST_PC(2);
	    }

	case INST_TRY_CVT_TO_NUMERIC:
	    {
		/*
		 * Try to convert the topmost stack object to an int or
		 * double object. This is done in order to support Tcl's
		 * policy of interpreting operands if at all possible as
		 * first integers, else floating-point numbers.
		 */
		
		double d;
		char *s;
		int converted, shared;

		valuePtr = stackPtr[stackTop].o;
		converted = 0;
		if ((valuePtr->typePtr != &tclIntType)
		        && (valuePtr->typePtr != &tclDoubleType)) {
		    s = TclGetStringFromObj(valuePtr, &length);
		    if (TclLooksLikeInt(s)) { /* FAILS IF NULLS */
			result = Tcl_GetIntFromObj(interp, valuePtr, &i);
		    } else {
			result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
		    }
		    if (result == TCL_OK) {
			converted = 1;
		    }
		    result = TCL_OK; /* reset the result variable */
		}

		/*
		 * Ensure that the topmost stack object, if numeric, has a
		 * string rep the same as the formatted version of its
		 * internal rep. This is used, e.g., to make sure that "expr
		 * {0001}" yields "1", not "0001". We implement this by
		 * _discarding_ the string rep since we know it will be
		 * regenerated, if needed later, by formatting the internal
		 * rep's value. Also check if there has been an IEEE
		 * floating point error.
		 */

		if ((valuePtr->typePtr == &tclIntType)
		        || (valuePtr->typePtr == &tclDoubleType)) {
		    shared = 0;
		    if (Tcl_IsShared(valuePtr)) {
			shared = 1;
			TclNewObj(objPtr);
			if (valuePtr->typePtr == &tclIntType) {
			    objPtr->internalRep.intValue =
				valuePtr->internalRep.intValue;
			    objPtr->typePtr = &tclIntType;
			} else {
			    objPtr->internalRep.doubleValue =
				valuePtr->internalRep.doubleValue;
			    objPtr->typePtr = &tclDoubleType;
			}
			TclDecrRefCount(valuePtr); /* done with object */
			valuePtr = objPtr;
		    } else {
			if (valuePtr->bytes != NULL) {
			    ckfree((char *) valuePtr->bytes);
			    valuePtr->bytes = NULL;
			}
		    }
		    stackPtr[stackTop].o = valuePtr; /* has ok ref ct */
		
		    if (valuePtr->typePtr == &tclDoubleType) {
			d = valuePtr->internalRep.doubleValue;
			if (IS_NAN(d) || IS_INF(d)) {
			    TRACE(("tryCvtToNumeric \"%.20s\" => IEEE FLOATING PT ERROR\n",
			           O2S(valuePtr)));
			    TclExprFloatError(interp, d);
			    result = TCL_ERROR;
			    goto checkForCatch;
			}
		    }
		    TRACE(("tryCvtToNumeric \"%.20s\" => numeric, %s, %s\n",
			   O2S(valuePtr),
			   (converted? "converted" : "not converted"),
			   (shared? "shared" : "not shared")));
		} else {
		    TRACE(("tryCvtToNumeric \"%.20s\" => not numeric\n",
			   O2S(valuePtr)));
		}
	    }
	    ADJUST_PC(1);

	case INST_BREAK:
	    /* Find the closest enclosing loop or catch exception range,
	     * if any. If a loop is found, terminate its execution.
	     * If the closest is a catch exception range, jump to its
	     * catchOffset. If no enclosing range is found, stop
	     * execution and return TCL_BREAK.
	     */
	
	    rangePtr = TclGetExceptionRangeForPc(pc, /*catchOnly*/ 0,
		    codePtr);
	    if (rangePtr == NULL) {
		TRACE(("break => no encl. loop or catch, returning TCL_BREAK\n"));
		result = TCL_BREAK;
		goto abnormalReturn; /* no catch exists to check */
	    }
	    switch (rangePtr->type) {
	    case LOOP_EXCEPTION_RANGE:
		result = TCL_OK;
		TRACE(("break => range at %d, new pc %d\n",
		       rangePtr->codeOffset, rangePtr->breakOffset));
		break;
	    case CATCH_EXCEPTION_RANGE:
		result = TCL_BREAK;
		TRACE(("break => ...\n"));
		goto processCatch; /* it will use rangePtr */
	    default:
		panic("TclExecuteByteCode: unrecognized ExceptionRange type %d\n", rangePtr->type);
	    }
	    pc = (codePtr->codeStart + rangePtr->breakOffset);
	    continue;	/* restart outer instruction loop at pc */

	case INST_CONTINUE:
            /*
	     * Find the closest enclosing loop or catch exception range,
	     * if any. If a loop is found, skip to its next iteration.
	     * If the closest is a catch exception range, jump to its
	     * catchOffset. If no enclosing range is found, stop
	     * execution and return TCL_CONTINUE.
	     */
	
	    rangePtr = TclGetExceptionRangeForPc(pc, /*catchOnly*/ 0,
		    codePtr);
	    if (rangePtr == NULL) {
		TRACE(("continue => no encl. loop or catch, returning TCL_CONTINUE\n"));
		result = TCL_CONTINUE;
		goto abnormalReturn; /* no catch exists to check */
	    }
	    switch (rangePtr->type) {
	    case LOOP_EXCEPTION_RANGE:
		if (rangePtr->continueOffset == -1) {
		    TRACE(("continue => loop w/o continue, checking for catch\n"));
		    goto checkForCatch;
		} else {
		    result = TCL_OK;
		    TRACE(("continue => range at %d, new pc %d\n",
			   rangePtr->codeOffset, rangePtr->continueOffset));
		}
		break;
	    case CATCH_EXCEPTION_RANGE:
		result = TCL_CONTINUE;
		TRACE(("continue => ...\n"));
		goto processCatch; /* it will use rangePtr */
	    default:
		panic("TclExecuteByteCode: unrecognized ExceptionRange type %d\n", rangePtr->type);
	    }
	    pc = (codePtr->codeStart + rangePtr->continueOffset);
	    continue;	/* restart outer instruction loop at pc */

	case INST_FOREACH_START4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    {
	        /*
		 * Initialize the temporary local var that holds the count
		 * of the number of iterations of the loop body to -1.
		 */

		ForeachInfo *infoPtr = (ForeachInfo *)
		    codePtr->auxDataArrayPtr[opnd].clientData;
		int iterTmpIndex = infoPtr->loopIterNumTmp;
		CallFrame *varFramePtr = iPtr->varFramePtr;
		Var *compiledLocals = varFramePtr->compiledLocals;
		Var *iterVarPtr;
		Tcl_Obj *oldValuePtr;

		iterVarPtr = &(compiledLocals[iterTmpIndex]);
#ifdef TCL_COMPILE_DEBUG		
		if (TclIsVarLink(iterVarPtr)) {
		    panic("TclExecuteByteCode execution failure: foreach loop iter temp %d is link\n", iterTmpIndex);
		}
#endif /* TCL_COMPILE_DEBUG */		
		oldValuePtr = iterVarPtr->value.objPtr;
#ifdef TCL_COMPILE_DEBUG		
		if ((oldValuePtr != NULL) && Tcl_IsShared(oldValuePtr)) {
		    panic("TclExecuteByteCode execution failure: foreach loop iter temp %d has shared object\n", iterTmpIndex);
		}
#endif /* TCL_COMPILE_DEBUG */		
		if (oldValuePtr == NULL) {
		    TclNewObj(objPtr);
		    objPtr->internalRep.intValue = -1;
		    objPtr->typePtr = &tclIntType;

		    iterVarPtr->value.objPtr = objPtr;
		    if (oldValuePtr != NULL) {
			TclDecrRefCount(oldValuePtr); /* discard old value */
		    }
		} else {	/* update object in place */
		    Tcl_ObjType *typePtr = oldValuePtr->typePtr;
		    if ((typePtr != NULL)
			    && (typePtr->freeIntRepProc != NULL)) {
			typePtr->freeIntRepProc(oldValuePtr);
		    }
		    oldValuePtr->typePtr = &tclIntType;
		    oldValuePtr->internalRep.intValue = -1;
		    if (oldValuePtr->bytes != NULL) {
			ckfree((char *) oldValuePtr->bytes);
			oldValuePtr->bytes = NULL;
		    }
		}
		TclSetVarScalar(iterVarPtr);
		TclClearVarUndefined(iterVarPtr);
		TRACE(("foreach_start4 %u => loop iter count temp %d\n", 
		        opnd, iterTmpIndex));
	    }
	    ADJUST_PC(5);
	
	case INST_FOREACH_STEP4:
	    opnd = TclGetUInt4AtPc(pc+1);
	    {
	        /*
		 * "Step" a foreach loop (i.e., begin its next iteration) by
		 * assigning the next value list element to each loop var.
		 */

		ForeachInfo *infoPtr = (ForeachInfo *)
		    codePtr->auxDataArrayPtr[opnd].clientData;
		ForeachVarList *varListPtr;
		int numLists = infoPtr->numLists;
		int iterTmpIndex = infoPtr->loopIterNumTmp;
		CallFrame *varFramePtr = iPtr->varFramePtr;
		Var *compiledLocals = varFramePtr->compiledLocals;
		int iterNum, listTmpIndex, listLen, numVars;
		int varIndex, valIndex, j;
		Tcl_Obj *listPtr, *elemPtr, *oldValuePtr;
		List *listRepPtr;
		Var *iterVarPtr, *listVarPtr;
		int continueLoop = 0;

		/*
		 * Increment the temp holding the loop iteration number.
		 */

		iterVarPtr = &(compiledLocals[iterTmpIndex]);
#ifdef TCL_COMPILE_DEBUG
		if (TclIsVarLink(iterVarPtr) || TclIsVarUndefined(iterVarPtr)
		        || !TclIsVarScalar(iterVarPtr)) {
		    panic("TclExecuteByteCode execution failure: foreach loop iter temp %d is link, undefined, or array\n", iterTmpIndex);
		}
#endif /* TCL_COMPILE_DEBUG */
		oldValuePtr = iterVarPtr->value.objPtr;
#ifdef TCL_COMPILE_DEBUG
		if ((oldValuePtr == NULL)
		        || (oldValuePtr->typePtr != &tclIntType)
		        || (oldValuePtr->bytes != NULL)
		        || Tcl_IsShared(oldValuePtr)) {
		    panic("TclExecuteByteCode execution failure: foreach loop iter count object is bad\n");
		}
		if ((oldValuePtr->typePtr != NULL)
		        && (oldValuePtr->typePtr->freeIntRepProc != NULL)) {
		    oldValuePtr->typePtr->freeIntRepProc(oldValuePtr);
		}
#endif /* TCL_COMPILE_DEBUG */
		oldValuePtr->internalRep.intValue++;
		iterNum = oldValuePtr->internalRep.intValue;
		oldValuePtr->typePtr = &tclIntType;
		
		/*
		 * Check whether all value lists are exhausted and we should
		 * stop the loop.
		 */

		listTmpIndex = infoPtr->firstListTmp;
		for (i = 0;  i < numLists;  i++) {
		    varListPtr = infoPtr->varLists[i];
		    numVars = varListPtr->numVars;

		    listVarPtr = &(compiledLocals[listTmpIndex]);
#ifdef TCL_COMPILE_DEBUG
		    if (TclIsVarLink(listVarPtr) || TclIsVarUndefined(listVarPtr)
		        || !TclIsVarScalar(listVarPtr)) {
			panic("TclExecuteByteCode execution failure: foreach loop list temp %d is link, undefined, or array\n", listTmpIndex);
		    }
#endif /* TCL_COMPILE_DEBUG */
		    listPtr = listVarPtr->value.objPtr;
#ifdef TCL_COMPILE_DEBUG
		    if (listPtr == NULL) {
			panic("TclExecuteByteCode execution failure: NULL foreach list temp %d: \"%s\"\n",
			      listTmpIndex,
			      TclGetStringFromObj(iPtr->objResult, &length));
		    }
#endif /* TCL_COMPILE_DEBUG */
		    if (listPtr->typePtr != &tclListType) {
			result = tclListType.setFromAnyProc(interp, listPtr);
			if (result != TCL_OK) {
			    TRACE_WITH_OBJ(("foreach_step4 %u => ERROR converting list %ld, \"%s\": ",
			            opnd, i, O2S(listPtr)),
				    iPtr->objResult);
			    goto checkForCatch;
			}
		    }
		    listRepPtr = (List *)listPtr->internalRep.otherValuePtr;
		    listLen = listRepPtr->elemCount;
		    
		    if (listLen > (iterNum * numVars)) {
			continueLoop = 1;
		    }
		    listTmpIndex++;
		}

		/*
		 * If some var in some var list still has a remaining list
		 * element iterate one more time. Assign to var the next
		 * element from its value list. We already checked above
		 * that each list temp holds a valid list object.
		 */
		
		if (continueLoop) {
		    listTmpIndex = infoPtr->firstListTmp;
		    for (i = 0;  i < numLists;  i++) {
			varListPtr = infoPtr->varLists[i];
			numVars = varListPtr->numVars;

			listVarPtr = &(compiledLocals[listTmpIndex]);
			listPtr = listVarPtr->value.objPtr;
			listRepPtr = (List *)
			        listPtr->internalRep.otherValuePtr;
			listLen = listRepPtr->elemCount;
			
			valIndex = (iterNum * numVars);
			for (j = 0;  j < numVars;  j++) {
			    int setEmptyStr = 0;
			    if (valIndex >= listLen) {
				setEmptyStr = 1;
				TclNewObj(elemPtr);  /* set to "" */
			    } else {
				elemPtr = listRepPtr->elements[valIndex];
			    }
			    
			    varIndex = varListPtr->varIndexes[j];
			    DECACHE_STACK_INFO();
			    value2Ptr = TclSetIndexedScalar(interp,
			           varIndex, elemPtr, /*leaveErrorMsg*/ 1);
			    CACHE_STACK_INFO();
			    if (setEmptyStr) {
				TclDecrRefCount(elemPtr);
			    }
			    if (value2Ptr == NULL) {
				TRACE_WITH_OBJ(("foreach_step4 %u => ERROR init. index temp %d: ",
				       opnd, varIndex), iPtr->objResult);
				result = TCL_ERROR;
				goto checkForCatch;
			    }
			    valIndex++;
			}
			listTmpIndex++;
		    }
		}
		
		/*
		 * Now push a "1" object if at least one value list had a
		 * remaining element and the loop should continue.
		 * Otherwise push "0".
		 */
		
		TclNewObj(objPtr);
		objPtr->typePtr = &tclIntType;
		objPtr->internalRep.intValue = continueLoop;
		stackPtr[++stackTop].o = objPtr; /* already has ref ct==1 */
		TRACE(("foreach_step4 %u => %d lists, iter %d, %s loop\n", 
		        opnd, numLists, iterNum,
		        (continueLoop? "continue" : "exit")));
	    }
	    ADJUST_PC(5);

	case INST_BEGIN_CATCH4:
	    /*
	     * Record start of the catch command with exception range index
	     * equal to the operand. Push the current stack depth onto the
	     * special catch stack.
	     */
	    catchStackPtr[++catchTop] = stackTop;
	    TRACE(("beginCatch4 %u => catchTop=%d, stackTop=%d\n",
		    TclGetUInt4AtPc(pc+1), catchTop, stackTop));
	    ADJUST_PC(5);

	case INST_END_CATCH:
	    catchTop--;
	    result = TCL_OK;
	    TRACE(("endCatch => catchTop=%d\n", catchTop));
	    ADJUST_PC(1);

	case INST_PUSH_RESULT:
	    valuePtr = iPtr->objResult;
	    Tcl_IncrRefCount(valuePtr);  /* since stack points to obj */
	    stackPtr[++stackTop].o = valuePtr;
	    TRACE_WITH_OBJ(("pushResult => "), valuePtr);
	    ADJUST_PC(1);

	case INST_PUSH_RETURN_CODE:
	    TclNewObj(objPtr);
	    objPtr->internalRep.intValue = result; /* i.e., return code */
	    objPtr->typePtr = &tclIntType;
	    stackPtr[++stackTop].o = objPtr;
	    TRACE_WITH_OBJ(("pushReturnCode => "), objPtr);
	    ADJUST_PC(1);

	default:
	    TRACE(("UNRECOGNIZED INSTRUCTION %u\n", opCode));
	    panic("TclExecuteByteCode: unrecognized opCode %u", opCode);
	} /* end of switch on opCode */

	/*
	 * Division by zero in an expression. Control only reaches this
	 * point by "goto divideByZero".
	 */
	
        divideByZero:
	TclSetObjResultFromString(interp, "divide by zero", -1);
	Tcl_SetErrorCode(interp, "ARITH", "DIVZERO", "divide by zero",
			 (char *) NULL);
	result = TCL_ERROR;
	
	/*
	 * Execution has generated an "exceptional return" (or "exception")
	 * such as TCL_ERROR. Look for the closest enclosing catch exception
	 * range, if any. If no enclosing catch range is found, stop
	 * execution and return the "exceptional return" code.
	 */
	
        checkForCatch:
	rangePtr = TclGetExceptionRangeForPc(pc, /*catchOnly*/ 1, codePtr);
	if (rangePtr == NULL) {
	    TRACE(("   ... no enclosing catch, returning %s\n",
		    StringForResultCode(result)));
	    goto abnormalReturn; /* no catch exists to check */
	}

	/*
	 * A catch exception range (rangePtr) has been to handle an
	 * "exception". It was found either by checkForCatch just above or
	 * by an instruction during break, continue, or error processing.
	 * Jump to its catchOffset after unwinding the operand stack to
	 * the depth it had when starting to execute the range's catch
	 * command. Also, if the exception is an error, record information
	 * about what was being executed when the error occurred.
	 */

        processCatch:
	if ((result == TCL_ERROR) && !(iPtr->flags & ERR_ALREADY_LOGGED)) {
	    char buf[200];
	    int cmdIndex = TclGetSrcInfoForPc(pc, codePtr);
	    
	    /*
	     * Compute the line number where the error occurred.
	     */
	    
	    iPtr->errorLine = 1; /* no correct line # information yet */

	    /*
	     * Print the command in the error message (up to a certain
	     * number of characters, or up to the first new-line).
	     */
	    
	    if (cmdIndex != -1) {
		CmdLocation *locPtr = &(codePtr->cmdMapPtr[cmdIndex]);
		char *ellipsis = "";
		int numChars = locPtr->numSrcChars;
		if (numChars > 150) {
		    numChars = 150;
		    ellipsis = "...";
		}
		if (!(iPtr->flags & ERR_IN_PROGRESS)) {
		    sprintf(buf, "\n    while executing\n\"%.*s%s\"",
			    numChars, (codePtr->source + locPtr->srcOffset),
			    ellipsis);
		} else {
		    sprintf(buf, "\n    invoked from within\n\"%.*s%s\"",
			    numChars, (codePtr->source + locPtr->srcOffset),
			    ellipsis);
		}
		Tcl_AddObjErrorInfo(interp, buf, -1);
		iPtr->flags |= ERR_ALREADY_LOGGED;
	    }
	}
	
	while (stackTop > catchStackPtr[catchTop]) {
	    valuePtr = POP_OBJECT();
	    TclDecrRefCount(valuePtr);
	}
	TRACE(("  ... found catch at %d, catchTop=%d, unwound to %d, new pc %u\n",
	        rangePtr->codeOffset, catchTop, catchStackPtr[catchTop],
	        (unsigned int)(rangePtr->catchOffset)));
	pc = (codePtr->codeStart + rangePtr->catchOffset);
	continue;		/* restart the execution loop at pc */
    } /* end of infinite loop dispatching on instructions */

    /*
     * Abnormal return code. Restore the stack to state it had when starting
     * to execute the ByteCode.
     */

    abnormalReturn:
    while (stackTop > initStackTop) {
	valuePtr = POP_OBJECT();
	TclDecrRefCount(valuePtr);
    }

    /*
     * Free the catch stack array if malloc'ed storage was used.
     */

    done:
    if (catchStackPtr != catchStackStorage) {
	ckfree((char *) catchStackPtr);
    }
    eePtr->stackTop = initStackTop;
    return result;
#undef STATIC_CATCH_STACK_SIZE
}

/*
 *----------------------------------------------------------------------
 *
 * IllegalExprOperandType --
 *
 *	Used by TclExecuteByteCode to add an error message to errorInfo
 *	when an illegal operand type is detected by an expression
 *	instruction. The argument opCode holds the failing instruction's
 *	opcode and opndPtr holds the operand object in error.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	An error message is appended to errorInfo.
 *
 *----------------------------------------------------------------------
 */

static void
IllegalExprOperandType(interp, opCode, opndPtr)
    Tcl_Interp *interp;		/* Interpreter to which error information
				 * pertains. */
    unsigned int opCode;	/* The instruction opcode being executed
				 * when the illegal type was found. */
    Tcl_Obj *opndPtr;		/* Points to the operand holding the value
				 * with the illegal type. */
{
    register Tcl_Obj *resultPtr;

    Tcl_ResetObjResult(interp);
    resultPtr = ((Interp *) interp)->objResult;
    if ((opndPtr->bytes == NULL) || (opndPtr->length == 0)) {
        Tcl_StringObjAppend(resultPtr,
		 "can't use empty string as operand of \"", -1);
        Tcl_StringObjAppend(resultPtr,
		 operatorStrings[opCode - INST_BITOR], -1);
        Tcl_StringObjAppend(resultPtr,
                 "\" (Note: expr no longer does a second round of substitutions.)", -1);
    } else {
	Tcl_StringObjAppend(resultPtr, "can't use ", -1);
        Tcl_StringObjAppend(resultPtr,
		    ((opndPtr->typePtr == &tclDoubleType) ?
		        "floating-point value" : "non-numeric string"), -1);
        Tcl_StringObjAppend(resultPtr,
                  " as operand of \"", -1);
	Tcl_StringObjAppend(resultPtr,
		  operatorStrings[opCode - INST_BITOR], -1);
        Tcl_StringObjAppend(resultPtr, "\"", -1);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * CallTraceProcedure --
 *
 *	Invokes a trace procedure registered with an interpreter. These
 *	procedures trace command execution. Currently this trace procedure
 *	is called with the address of the string-based Tcl_CmdProc for the
 *	command, not the Tcl_ObjCmdProc.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Those side effects made by the trace procedure.
 *
 *----------------------------------------------------------------------
 */

static void
CallTraceProcedure(interp, tracePtr, cmdPtr, objc, objv)
    Tcl_Interp *interp;		/* The current interpreter. */
    register Trace *tracePtr;	/* Describes the trace procedure to call. */
    Command *cmdPtr;		/* Points to command's Command struct. */
    register int objc;		/* Number of arguments for the command. */
    Tcl_Obj *objv[];		/* Pointers to Tcl_Obj of each argument. */
{
    Interp *iPtr = (Interp *) interp;
    register char **argv;
    register int i;
    int length;

    /*
     * Get the string rep from the objv argument objects and place their
     * pointers in argv. First make sure objv is large enough to hold the
     * objc args plus 1 extra for the zero end-of-argv word.
     * THIS FAILS IF AN OBJECT'S STRING REP CONTAINS NULLS.
     */
    
    argv = (char **) ckalloc((unsigned)(objc + 1) * sizeof(char *));
    for (i = 0;  i < objc;  i++) {
	argv[i] = TclGetStringFromObj(objv[i], &length);
    }
    argv[objc] = 0;

    /*
     * Call the trace procedure.
     */
    
    (*tracePtr->proc)(tracePtr->clientData, interp, iPtr->numLevels,
                      /*cmdStart*/ 0, cmdPtr->proc, cmdPtr->clientData,
		      objc, argv);

    ckfree((char *) argv);
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetSrcInfoForPc --
 *
 *	Procedure that given a program counter value, returns an index
 *	of the closest command's element in the bytecode code unit's
 *	CmdLocation array. This element provides information about that
 *	command's source: a pointer to its first byte and the number
 *	of its characters.
 *
 * Results:
 *	If a command in the bytecode code unit is found that encloses
 *	the program counter value, the index of the command's element
 *	in the CmdLocation array is returned. If multiple commands
 *	resulted in code at pc, the index for the command with code that
 *	starts closest to pc is returned. If no matching command is
 *	found, -1 is returned.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

int
TclGetSrcInfoForPc(pc, codePtr)
    unsigned char *pc;		/* The program counter value for which to
				 * return the closest command's source info.
				 * This points to a bytecode instruction
				 * in codePtr's code. */
    ByteCode* codePtr;		/* The bytecode sequence in which to look
				 * up the command source for the pc. */
{
    int codeOffset = (pc - codePtr->codeStart);
    int numCommands = codePtr->numCommands;
    CmdLocation *cmdMapPtr = codePtr->cmdMapPtr;
    register CmdLocation *locPtr;
    int bestCmd = -1;		/* Index of current candidate for closest
				 * command. */
    int bestDist = INT_MAX;	/* Distance of pc to bestCmd's start pc. */
    int startOffset, endOffset, dist;
    register int i;

    for (i = 0;  i < numCommands;  i++) {
	locPtr = &(cmdMapPtr[i]);
	startOffset = locPtr->codeOffset;
	endOffset = (startOffset + locPtr->numCodeBytes - 1);
	if ((startOffset <= codeOffset) && (codeOffset <= endOffset)) {
	    dist = (codeOffset - startOffset);
	    if (dist <= bestDist) {
		bestCmd = i;
		bestDist = dist;
	    }
	}
    }
    return bestCmd;
}

/*
 *----------------------------------------------------------------------
 *
 * TclGetExceptionRangeForPc --
 *
 *	Procedure that given a program counter value, returns the closest
 *	enclosing ExceptionRange that matches the kind requested.
 *
 * Results:
 *	In the normal case, catchOnly is 0 (false) and this procedure
 *	returns a pointer to the most closely enclosing ExceptionRange
 *	structure regardless of whether it is a loop or catch exception
 *	range. This is appropriate when processing a TCL_BREAK or
 *	TCL_CONTINUE, which will be "handled" either by a loop exception
 *	range or a closer catch range. If catchOnly is nonzero (true), this
 *	procedure ignores loop exception ranges and returns a pointer to the
 *	closest catch range. If no matching ExceptionRange is found that
 *	encloses pc, a NULL is returned.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

ExceptionRange *
TclGetExceptionRangeForPc(pc, catchOnly, codePtr)
    unsigned char *pc;		/* The program counter value for which to
				 * search for a closest enclosing exception
				 * range. This points to a bytecode
				 * instruction in codePtr's code. */
    int catchOnly;		/* If 0, consider either loop or catch
				 * ExceptionRanges in search. Otherwise
				 * consider only catch ranges (and ignore
				 * any closer loop ranges). */
    ByteCode* codePtr;		/* Points to the ByteCode in which to search
				 * for the enclosing ExceptionRange. */
{
    ExceptionRange *rangeArrayPtr = codePtr->excRangeArrayPtr;
    int numRanges = codePtr->numExcRanges;
    register ExceptionRange *rangePtr;
    int codeOffset = (pc - codePtr->codeStart);
    register int i, level;

    for (level = codePtr->maxExcRangeDepth;  level >= 0;  level--) {
	for (i = 0;  i < numRanges;  i++) {
	    rangePtr = &(rangeArrayPtr[i]);
	    if (rangePtr->nestingLevel == level) {
		int start = rangePtr->codeOffset;
		int end   = (start + rangePtr->numCodeBytes);
		if ((start <= codeOffset) && (codeOffset < end)) {
		    if ((!catchOnly)
			    || (rangePtr->type == CATCH_EXCEPTION_RANGE)) {
			return rangePtr;
		    }
		}
	    }
	}
    }
    return NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Math Functions --
 *
 *	This page contains the procedures that implement all of the
 *	built-in math functions for expressions.
 *
 * Results:
 *	Each procedure returns TCL_OK if it succeeds and pushes an
 *	Tcl object holding the result. If it fails it returns TCL_ERROR
 *	and leaves an error message in interp->objResult.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
ExprUnaryFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes one double argument and returns a
				 * double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *objPtr;
    double d, dResult;
    long i;
    int length;
    char *s;
    int result = TCL_OK;
    
    double (*func) _ANSI_ARGS_((double)) =
	(double (*)_ANSI_ARGS_((double))) clientData;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the function's argument from the evaluation stack. Convert it
     * to a double if necessary.
     */

    valuePtr = POP_OBJECT();
    if (valuePtr->typePtr == &tclIntType) {
	d = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	d = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		d = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		d = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
	            "argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);  /* done with popped obj */

    errno = 0;
    dResult = (*func)(d);
    if ((errno != 0) || IS_NAN(dResult) || IS_INF(dResult)) {
	TclExprFloatError(interp, dResult);
	result = TCL_ERROR;
	goto done;
    }
    
    /*
     * Push a Tcl object holding the result.
     */

    TclNewObj(objPtr);
    objPtr->internalRep.doubleValue = dResult;
    objPtr->typePtr = &tclDoubleType;
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */
    
    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

static int
ExprBinaryFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes two double arguments and
				 * returns a double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *value2Ptr, *objPtr;
    double d1, d2, dResult;
    long i;
    int length;
    char *s;
    int result = TCL_OK;
    
    double (*func) _ANSI_ARGS_((double, double))
	= (double (*)_ANSI_ARGS_((double, double))) clientData;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the function's two arguments from the evaluation stack. Convert
     * them to doubles if necessary.
     */

    value2Ptr = POP_OBJECT();
    valuePtr  = POP_OBJECT();
    
    if (valuePtr->typePtr == &tclIntType) {
	d1 = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	d1 = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &d1);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		d1 = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		d1 = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
	            "argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);   /* done with popped obj */
	    TclDecrRefCount(value2Ptr);  /* done with popped obj */
	    goto done;
	}
    }
	
    if (value2Ptr->typePtr == &tclIntType) {
	d2 = value2Ptr->internalRep.intValue;
    } else if (value2Ptr->typePtr == &tclDoubleType) {
	d2 = value2Ptr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(value2Ptr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, value2Ptr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, value2Ptr, &d2);
	}
	if (result == TCL_OK) {
	    if (value2Ptr->typePtr == &tclIntType) {
		d2 = value2Ptr->internalRep.intValue;
	    } else if (value2Ptr->typePtr == &tclDoubleType) {
		d2 = value2Ptr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
	            "argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);   /* done with popped obj */
	    TclDecrRefCount(value2Ptr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);   /* done with popped obj */
    TclDecrRefCount(value2Ptr);  /* done with popped obj */

    errno = 0;
    dResult = (*func)(d1, d2);
    if ((errno != 0) || IS_NAN(dResult) || IS_INF(dResult)) {
	TclExprFloatError(interp, dResult);
	result = TCL_ERROR;
	goto done;
    }

    /*
     * Push a Tcl object holding the result.
     */

    TclNewObj(objPtr);
    objPtr->internalRep.doubleValue = dResult;
    objPtr->typePtr = &tclDoubleType;
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */
    
    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

static int
ExprAbsFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes two double arguments and
				 * returns a double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *objPtr;
    long i, iResult;
    double d, dResult;
    int haveInt = 0;		/* Initialized to avoid compiler warning. */
    int result = TCL_OK;
    int length;
    char *s;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the argument from the evaluation stack.
     */

    valuePtr = POP_OBJECT();
    if (valuePtr->typePtr == &tclIntType) {
	haveInt = 1;
	i = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	haveInt = 0;
	d = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		haveInt = 1;
		i = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		haveInt = 0;
		d = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
	        "argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);  /* done with popped obj */

    /*
     * Push a Tcl object with the result.
     */
    
    if (haveInt) {
	if (i < 0) {
	    iResult = -i;
	    if (iResult < 0) {
		TclSetObjResultFromString(interp,
		        "integer value too large to represent", -1);
		Tcl_SetErrorCode(interp, "ARITH", "IOVERFLOW",
			"integer value too large to represent",
			(char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	} else {
	    iResult = i;
	}	    
	TclNewObj(objPtr);
	objPtr->internalRep.intValue = iResult;
	objPtr->typePtr = &tclIntType;
	stackPtr[++stackTop].o = objPtr; /* already has correct ref count */
    } else {
	if (d < 0.0) {
	    dResult = -d;
	} else {
	    dResult = d;
	}
	if (IS_NAN(dResult) || IS_INF(dResult)) {
	    TclExprFloatError(interp, dResult);
	    result = TCL_ERROR;
	    goto done;
	}
	TclNewObj(objPtr);
	objPtr->internalRep.doubleValue = dResult;
	objPtr->typePtr = &tclDoubleType;
	stackPtr[++stackTop].o = objPtr; /* already has correct ref count */
    }
    
    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

static int
ExprDoubleFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes two double arguments and
				 * returns a double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *objPtr;
    double dResult;
    long i;
    int length;
    char *s;
    int result = TCL_OK;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the argument from the evaluation stack.
     */

    valuePtr = POP_OBJECT();
    if (valuePtr->typePtr == &tclIntType) {
	dResult = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	dResult = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &dResult);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		dResult = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		dResult = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
		    "argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);  /* done with popped obj */

    /*
     * Push a Tcl object with the result.
     */

    TclNewObj(objPtr);
    objPtr->internalRep.doubleValue = dResult;
    objPtr->typePtr = &tclDoubleType;
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */

    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

static int
ExprIntFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes two double arguments and
				 * returns a double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *objPtr;
    long i = 0;			/* Initialized to avoid compiler warning. */
    long iResult;
    double d;
    int haveInt = 0;		/* Initialized to avoid compiler warning. */
    int length;
    char *s;
    int result = TCL_OK;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the argument from the evaluation stack.
     */

    valuePtr = POP_OBJECT();
    if (valuePtr->typePtr == &tclIntType) {
	haveInt = 1;
	i = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	haveInt = 0;
	d = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		haveInt = 1;
		i = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		haveInt = 0;
		d = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
		"argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);  /* done with popped obj */

    /*
     * Push a Tcl object with the result.
     */
    
    if (haveInt) {
	iResult = i;
    } else {
	if (d < 0.0) {
	    if (d < (double) (long) LONG_MIN) {
		tooLarge:
		TclSetObjResultFromString(interp,
		        "integer value too large to represent", -1);
		Tcl_SetErrorCode(interp, "ARITH", "IOVERFLOW",
			"integer value too large to represent",
			(char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	} else {
	    if (d > (double) LONG_MAX) {
		goto tooLarge;
	    }
	}
	if (IS_NAN(d) || IS_INF(d)) {
	    TclExprFloatError(interp, d);
	    result = TCL_ERROR;
	    goto done;
	}
	iResult = (long) d;
    }
    TclNewObj(objPtr);
    objPtr->internalRep.intValue = iResult;
    objPtr->typePtr = &tclIntType;
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */

    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

static int
ExprRoundFunc(interp, eePtr, clientData)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    ClientData clientData;	/* Contains the address of a procedure that
				 * takes two double arguments and
				 * returns a double result. */
{
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    Tcl_Obj *valuePtr, *objPtr;
    long i = 0;			/* Initialized to avoid compiler warning. */
    long iResult;
    double d, temp;
    int haveInt = 0;		/* Initialized to avoid compiler warning. */
    int length;
    char *s;
    int result = TCL_OK;

    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Pop the argument from the evaluation stack.
     */

    valuePtr = POP_OBJECT();
    if (valuePtr->typePtr == &tclIntType) {
	haveInt = 1;
	i = valuePtr->internalRep.intValue;
    } else if (valuePtr->typePtr == &tclDoubleType) {
	haveInt = 0;
	d = valuePtr->internalRep.doubleValue;
    } else {
	s = TclGetStringFromObj(valuePtr, &length);
	if (TclLooksLikeInt(s)) {
	    result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	} else {
	    result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
	}
	if (result == TCL_OK) {
	    if (valuePtr->typePtr == &tclIntType) {
		haveInt = 1;
		i = valuePtr->internalRep.intValue;
	    } else if (valuePtr->typePtr == &tclDoubleType) {
		haveInt = 0;
		d = valuePtr->internalRep.doubleValue;
	    }
	} else {
	    TclSetObjResultFromString(interp,
		"argument to math function didn't have numeric value", -1);
	    TclDecrRefCount(valuePtr);  /* done with popped obj */
	    goto done;
	}
    }
    TclDecrRefCount(valuePtr);  /* done with popped obj */

    /*
     * Push a Tcl object with the result.
     */
    
    if (haveInt) {
	iResult = i;
    } else {
	if (d < 0.0) {
	    if (d <= (((double) (long) LONG_MIN) - 0.5)) {
		tooLarge:
		TclSetObjResultFromString(interp,
		        "integer value too large to represent", -1);
		Tcl_SetErrorCode(interp, "ARITH", "IOVERFLOW",
			"integer value too large to represent",
			(char *) NULL);
		result = TCL_ERROR;
		goto done;
	    }
	    temp = (long) (d - 0.5);
	} else {
	    if (d >= (((double) LONG_MAX + 0.5))) {
		goto tooLarge;
	    }
	    temp = (long) (d + 0.5);
	}
	if (IS_NAN(temp) || IS_INF(temp)) {
	    TclExprFloatError(interp, temp);
	    result = TCL_ERROR;
	    goto done;
	}
	iResult = (long) temp;
    }
    TclNewObj(objPtr);
    objPtr->internalRep.intValue = iResult;
    objPtr->typePtr = &tclIntType;
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */

    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * ExprCallMathFunc --
 *
 *	This procedure is invoked to call a non-builtin math function
 *	during the execution of an expression. 
 *
 * Results:
 *	TCL_OK is returned if all went well and the function's value
 *	was computed successfully. If an error occurred, TCL_ERROR
 *	is returned and an error message is left in interp->objResult.
 *	After a successful return this procedure pushes a Tcl object
 *	holding the result. 
 *
 * Side effects:
 *	None, unless the called math function has side effects.
 *
 *----------------------------------------------------------------------
 */

static int
ExprCallMathFunc(interp, eePtr, objc, objv)
    Tcl_Interp *interp;		/* The interpreter in which to execute the
				 * function. */
    ExecEnv *eePtr;		/* Points to the environment for executing
				 * the function. */
    int objc;			/* Number of arguments. The function name is
				 * the 0-th argument. */
    Tcl_Obj **objv;		/* The array of arguments. The function name
				 * is objv[0]. */
{
    Interp *iPtr = (Interp *) interp;
    StackItem *stackPtr;        /* Cached evaluation stack base pointer. */
    register int stackTop;	/* Cached top index of evaluation stack. */
    char *funcName, *s;
    Tcl_HashEntry *hPtr;
    MathFunc *mathFuncPtr;	/* Information about math function. */
    Tcl_Value args[MAX_MATH_ARGS]; /* Arguments for function call. */
    Tcl_Value funcResult;	/* Result of function call as Tcl_Value. */
    Tcl_Obj *valuePtr, *objPtr;
    int haveInt = 1;		/* 1 if Tcl object for arg holds integer;
				 * otherwise 0. */
    long i;
    double d;
    int length, j, k, result;
    
    /*
     * Set stackPtr and stackTop from eePtr.
     */
    
    CACHE_STACK_INFO();

    /*
     * Look up the MathFunc record for the function.
     * THIS FAILS IF THE OBJECT'S STRING REP CONTAINS NULLS.
     */

    funcName = TclGetStringFromObj(objv[0], &length);
    hPtr = Tcl_FindHashEntry(&iPtr->mathFuncTable, funcName);
    if (hPtr == NULL) {
	Tcl_ResetObjResult(interp);
        Tcl_StringObjAppend(iPtr->objResult, "unknown math function \"", -1);
        Tcl_StringObjAppendObj(iPtr->objResult, objv[0]);
        Tcl_StringObjAppend(iPtr->objResult, "\"", -1);
	result = TCL_ERROR;
	goto done;
    }
    mathFuncPtr = (MathFunc *) Tcl_GetHashValue(hPtr);
    if (mathFuncPtr->numArgs != (objc-1)) {
	panic("ExprCallMathFunc: expected number of args %d != actual number %d",
	        mathFuncPtr->numArgs, objc);
	result = TCL_ERROR;
	goto done;
    }

    /*
     * Collect the arguments for the function, if there are any, into the
     * array "args". Note that args[0] will have the Tcl_Value that
     * corresponds to objv[1].
     */

    for (j = 1, k = 0;  j < objc;  j++, k++) {
	valuePtr = objv[j];
	if (valuePtr->typePtr == &tclIntType) {
	    haveInt = 1;
	    i = valuePtr->internalRep.intValue;
	} else if (valuePtr->typePtr == &tclDoubleType) {
	    haveInt = 0;
	    d = valuePtr->internalRep.doubleValue;
	} else {      	  /* try to convert to int first then double */
	    s = TclGetStringFromObj(valuePtr, &length);
	    if (TclLooksLikeInt(s)) {
		result = Tcl_GetIntFromObj(interp, valuePtr, &i);
	    } else {
		result = Tcl_GetDoubleFromObj(interp, valuePtr, &d);
	    }
	    if (result == TCL_OK) {
		if (valuePtr->typePtr == &tclIntType) {
		    haveInt = 1;
		    i = valuePtr->internalRep.intValue;
		} else if (valuePtr->typePtr == &tclDoubleType) {
		    haveInt = 0;
		    d = valuePtr->internalRep.doubleValue;
		}
	    } else {
		TclSetObjResultFromString(interp,
			"argument to math function didn't have numeric value", -1);
		goto done;
	    }
	}

	/*
	 * Copy the object's numeric value to the argument record,
	 * converting it if necessary. 
	 */
	
	if (haveInt) {
	    if (mathFuncPtr->argTypes[k] == TCL_DOUBLE) {
		args[k].type = TCL_DOUBLE;
		args[k].doubleValue = i;
	    } else {
		args[k].type = TCL_INT;
		args[k].intValue = i;
	    }
	} else {
	    if (mathFuncPtr->argTypes[k] == TCL_INT) {
		args[k].type = TCL_INT;
		args[k].intValue = (long) d;
	    } else {
		args[k].type = TCL_DOUBLE;
		args[k].doubleValue = d;
	    }
	}
    }

    /*
     * Invoke the function and copy its result back into valuePtr.
     */

    tcl_MathInProgress++;
    result = (*mathFuncPtr->proc)(mathFuncPtr->clientData, interp, args,
	    &funcResult);
    tcl_MathInProgress--;
    if (result != TCL_OK) {
	goto done;
    }

    /*
     * Pop the objc top stack elements and decrement their ref counts.
     */
		
    i = (stackTop - (objc-1));
    while (i <= stackTop) {
	valuePtr = stackPtr[i].o;
	TclDecrRefCount(valuePtr);
	i++;
    }
    stackTop -= objc;
    
    /*
     * Push the call's object result.
     */
    
    TclNewObj(objPtr);
    if (funcResult.type == TCL_INT) {
	objPtr->internalRep.intValue = funcResult.intValue;
	objPtr->typePtr = &tclIntType;
    } else {
	d = funcResult.doubleValue;
	if (IS_NAN(d) || IS_INF(d)) {
	    TclExprFloatError(interp, d);
	    result = TCL_ERROR;
	    goto done;
	}
	objPtr->internalRep.doubleValue = d;
	objPtr->typePtr = &tclDoubleType;
    }
    stackPtr[++stackTop].o = objPtr; /* already has correct ref count */

    /*
     * Reflect the change to stackTop back in eePtr.
     */

    done:
    DECACHE_STACK_INFO();
    return result;
}

/*
 *----------------------------------------------------------------------
 *
 * TclExprFloatError --
 *	This procedure is called when an error occurs during a
 *	floating-point operation. It reads errno and sets interp->objResult
 *	accordingly.
 *
 * Results:
 *	Interp->objResult is set to hold an error message.
 *
 * Side effects:
 *	None.
 *
 *---------------------------------------------------------------------- */

void
TclExprFloatError(interp, value)
    Tcl_Interp *interp;		/* Where to store error message. */
    double value;		/* Value returned after error;  used to
				 * distinguish underflows from overflows. */
{
    Interp *iPtr = (Interp *) interp;
    char *s;
    int length;

    if ((errno == EDOM) || (value != value)) {
	s = "domain error: argument not in valid range";
	TclSetObjResultFromString(interp, s, -1);
	Tcl_SetErrorCode(interp, "ARITH", "DOMAIN", s, (char *) NULL);
    } else if ((errno == ERANGE) || IS_INF(value)) {
	if (value == 0.0) {
	    s = "floating-point value too small to represent";
	    TclSetObjResultFromString(interp, s, -1);
	    Tcl_SetErrorCode(interp, "ARITH", "UNDERFLOW", s, (char *) NULL);
	} else {
	    s = "floating-point value too large to represent";
	    TclSetObjResultFromString(interp, s, -1);
	    Tcl_SetErrorCode(interp, "ARITH", "OVERFLOW", s, (char *) NULL);
	}
    } else {			/* FAILS IF STRING REP CONTAINS NULLS */
	char msg[100];
	
	sprintf(msg, "unknown floating-point error, errno = %d", errno);
	TclSetObjResultFromString(interp, msg, -1);
	Tcl_SetErrorCode(interp, "ARITH", "UNKNOWN",
	       TclGetStringFromObj(iPtr->objResult, &length),
	       (char *) NULL);
    }
}

#ifdef TCL_COMPILE_STATS
/*
 *----------------------------------------------------------------------
 *
 * EvalStatsCmd --
 *
 *	Implements the "evalstats" command that prints instruction execution
 *	counts to stdout.
 *
 * Results:
 *	Standard Tcl results.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
EvalStatsCmd(unused, interp, argc, argv)
    ClientData unused;		/* Unused. */
    Tcl_Interp *interp;		/* The current interpreter. */
    int argc;			/* The number of arguments. */
    char **argv;		/* The argument strings. */
{
    register double total = 0.0;
    register int i;

    for (i = 0;  i < 256;  i++) {
        if (instructionCount[i]) {
            total += instructionCount[i];
        }
    }   

    fprintf(stdout, "\nNumber of ByteCode compilations: %ld\n",
	    tclNumCompilations);
    fprintf(stdout, "Number of ByteCode executions: %ld\n",
	    numExecutions);
    fprintf(stdout, "Number of Tcl objects in use: %ld, allocated %ld, freed %ld\n",
	    (tclObjsAlloced - tclObjsFreed), tclObjsAlloced, tclObjsFreed);
    fprintf(stdout, "Number of instructions executed: %.0f\n\n", total);
    for (i = 0;  i < 256;  i++) {
        if (instructionCount[i]) {
            fprintf(stdout, "%30s %8d %6.2f%%\n",
		    opName[i], instructionCount[i],
		    (instructionCount[i] * 100.0)/total);
        }
    }

#ifdef TCL_MEM_DEBUG
    fprintf(stdout, "\nHeap Statistics:\n");
    TclDumpMemoryInfo(stdout);
#endif /* TCL_MEM_DEBUG */

    return TCL_OK;
}
#endif /* TCL_COMPILE_STATS */

/*
 *----------------------------------------------------------------------
 *
 * FreeCmdNameInternalRep --
 *
 *	Frees the resources associated with a cmdName object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Decrements the ref count of any cached Command structure pointed
 *	to by the cmdName's internal representation. This may free the
 *	Command structure.
 *
 *----------------------------------------------------------------------
 */

static void
FreeCmdNameInternalRep(objPtr)
    register Tcl_Obj *objPtr;	/* CmdName object with internal
				 * representation to free. */
{
    register Command *cmdPtr =
	(Command *) objPtr->internalRep.twoPtrValue.ptr1;

    /*
     * Now free the Command structure, unless it is still in the
     * interpreter's command hashtable or if there is another reference to
     * it from another CmdName object. This is an inline version of
     * TclCleanupCommand.
     */

    if (cmdPtr != NULL) {
	TclCleanupCommand(cmdPtr);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * DupCmdNameInternalRep --
 *
 *	Initialize the internal representation of an cmdName Tcl_Obj to a
 *	copy of the internal representation of an existing cmdName object. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"copyPtr"s internal rep is set to the cmdName corresponding to
 *	"srcPtr"s internal rep. Increments the ref count of any cached
 *	Command structure pointed to by the cmdName's internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static void
DupCmdNameInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    register Tcl_Obj *copyPtr;	/* Object with internal rep to set. */
{
    register Command *cmdPtr = (Command *)
	srcPtr->internalRep.twoPtrValue.ptr1;
    register Tcl_HashEntry *cachedHtblPtr = (Tcl_HashEntry *)
	srcPtr->internalRep.twoPtrValue.ptr2;

    copyPtr->internalRep.twoPtrValue.ptr1 = (VOID *) cmdPtr;
    copyPtr->internalRep.twoPtrValue.ptr2 = (VOID *) cachedHtblPtr;
    if (cmdPtr != NULL) {
	cmdPtr->refCount++;	/* since there is another reference */
    }
    copyPtr->typePtr = &tclCmdNameType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetCmdNameFromAny --
 *
 *	Generate an cmdName internal form for the Tcl object "objPtr".
 *
 * Results:
 *	The return value is a standard Tcl result. The conversion always
 *	succeeds and TCL_OK is returned.
 *
 * Side effects:
 *	A pointer to the Command structure for the command with name that
 *	matches objPtr's string rep is stored as objPtr's internal
 *	representation. This pointer will be NULL if no such command was
 *	found or if the command is not a core command. The ref count of
 *	any cached Command structure pointed to by the cmdName's internal
 *	representation is also incremented.
 *
 *----------------------------------------------------------------------
 */

static int
SetCmdNameFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    register Tcl_Obj *objPtr;	/* The object to convert. */
{
    Interp *iPtr = (Interp *) interp;
    register Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    char *name;
    register Tcl_HashEntry *hPtr; 
    Command *cmdPtr = NULL;
    int length;

    /*
     * Get "objPtr"s string representation. Make it up-to-date if necessary.
     * THIS FAILS IF THE OBJECT'S STRING REP CONTAINS NULLS.
     */

    name = TclGetStringFromObj(objPtr, &length);

    /*
     * Find the Command structure, if any, that describes the command called
     * "name". Increment the reference count in the Command structure to
     * reflect the new reference from the CmdName object and, if the command
     * is deleted later, to keep the Command structure from being freed
     * until TclExecuteByteCode has a chance to recognize that the command
     * was deleted.
     *
     * If the command procedure is TclObjInterpProc because the command is a
     * user-defined Tcl procedure, we instead cache the command's
     * objClientData (which points to the procedure's Proc structure) and
     * set the low order bit 1 to indicate that the cached "command
     * procedure" actually stores the objClientData.
     */

    hPtr = Tcl_FindHashEntry(&iPtr->commandTable, name);
    if (hPtr != NULL) {
	cmdPtr = (Command *) Tcl_GetHashValue(hPtr);
	cmdPtr->refCount++;
    }

    /*
     * If command wasn't found and the command name's bytes field is NULL
     * (e.g., if the command name is an empty string), change the
     * bytes field to point to an empty string on the heap. This ensures
     * that Tcl_GetStringFromObj won't later call UpdateStringOfCmdName
     * below, which only panics.
     */

    if ((cmdPtr == NULL) && (objPtr->bytes == NULL)) {
	objPtr->bytes = (char *) ckalloc((unsigned) 1);
	objPtr->bytes[0] = '\0';
	objPtr->length = 0;
    }

    /*
     * Free the old internalRep before setting the new one. We do this as
     * late as possible to allow the conversion code, in particular
     * GetStringFromObj, to use that old internalRep. If no Command
     * structure was found, leave NULL as the cached value.
     */

    if ((oldTypePtr != NULL) &&	(oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }
    
    objPtr->internalRep.twoPtrValue.ptr1 = (VOID *) cmdPtr;
    objPtr->internalRep.twoPtrValue.ptr2 = (VOID *) hPtr;
    objPtr->typePtr = &tclCmdNameType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfCmdName --
 *
 *	Update the string representation for an cmdName object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Generates a panic.  This procedure should never be invoked: the
 *	internal representation of a cmdName object is never modified,
 *	so the string representation should never be invalidated.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfCmdName(objPtr)
    Tcl_Obj *objPtr;		/* CmdName obj to update string rep. */
{
    panic("UpdateStringOfCmdName should never be invoked");
}

#ifdef TCL_COMPILE_DEBUG
/*
 *----------------------------------------------------------------------
 *
 * StringForResultCode --
 *
 *	Procedure that returns a human-readable string representing a
 *	Tcl result code such as TCL_ERROR. 
 *
 * Results:
 *	If the result code is one of the standard Tcl return codes, the
 *	result is a string representing that code such as "TCL_ERROR".
 *	Otherwise, the result string is that code formatted as a
 *	sequence of decimal digit characters. Note that the resulting
 *	string must not be modified by the caller.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static char *
StringForResultCode(result)
    int result;			/* The Tcl result code for which to
				 * generate a string. */
{
    static char buf[20];
    
    if ((result >= TCL_OK) && (result <= TCL_CONTINUE)) {
	return resultStrings[result];
    }
    TclFormatInt(buf, result);
    return buf;
}
#endif /* TCL_COMPILE_DEBUG */
